# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'rare.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QFormLayout,
    QGridLayout, QGroupBox, QHBoxLayout, QLabel,
    QPushButton, QSizePolicy, QSpacerItem, QVBoxLayout,
    QWidget)

class Ui_RareSettings(object):
    def setupUi(self, RareSettings):
        if not RareSettings.objectName():
            RareSettings.setObjectName(u"RareSettings")
        RareSettings.resize(629, 447)
        RareSettings.setWindowTitle(u"RareSettings")
        self.main_layout = QHBoxLayout(RareSettings)
        self.main_layout.setObjectName(u"main_layout")
        self.left_layout = QVBoxLayout()
        self.left_layout.setObjectName(u"left_layout")
        self.interface_group = QGroupBox(RareSettings)
        self.interface_group.setObjectName(u"interface_group")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.interface_group.sizePolicy().hasHeightForWidth())
        self.interface_group.setSizePolicy(sizePolicy)
        self.interface_layout = QFormLayout(self.interface_group)
        self.interface_layout.setObjectName(u"interface_layout")
        self.interface_layout.setFieldGrowthPolicy(QFormLayout.FieldsStayAtSizeHint)
        self.interface_layout.setLabelAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)
        self.lang_label = QLabel(self.interface_group)
        self.lang_label.setObjectName(u"lang_label")

        self.interface_layout.setWidget(0, QFormLayout.LabelRole, self.lang_label)

        self.lang_select = QComboBox(self.interface_group)
        self.lang_select.setObjectName(u"lang_select")

        self.interface_layout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.lang_select)

        self.color_label = QLabel(self.interface_group)
        self.color_label.setObjectName(u"color_label")

        self.interface_layout.setWidget(1, QFormLayout.LabelRole, self.color_label)

        self.color_select = QComboBox(self.interface_group)
        self.color_select.setObjectName(u"color_select")

        self.interface_layout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.color_select)

        self.style_label = QLabel(self.interface_group)
        self.style_label.setObjectName(u"style_label")

        self.interface_layout.setWidget(2, QFormLayout.LabelRole, self.style_label)

        self.style_select = QComboBox(self.interface_group)
        self.style_select.setObjectName(u"style_select")

        self.interface_layout.setWidget(2, QFormLayout.ItemRole.FieldRole, self.style_select)

        self.interface_info = QLabel(self.interface_group)
        self.interface_info.setObjectName(u"interface_info")
        font = QFont()
        font.setItalic(True)
        self.interface_info.setFont(font)
        self.interface_info.setWordWrap(True)

        self.interface_layout.setWidget(4, QFormLayout.SpanningRole, self.interface_info)

        self.view_label = QLabel(self.interface_group)
        self.view_label.setObjectName(u"view_label")

        self.interface_layout.setWidget(3, QFormLayout.LabelRole, self.view_label)

        self.view_combo = QComboBox(self.interface_group)
        self.view_combo.setObjectName(u"view_combo")

        self.interface_layout.setWidget(3, QFormLayout.ItemRole.FieldRole, self.view_combo)


        self.left_layout.addWidget(self.interface_group)

        self.settings_group = QGroupBox(RareSettings)
        self.settings_group.setObjectName(u"settings_group")
        self.behavior_layout = QGridLayout(self.settings_group)
        self.behavior_layout.setObjectName(u"behavior_layout")
        self.save_size = QCheckBox(self.settings_group)
        self.save_size.setObjectName(u"save_size")

        self.behavior_layout.addWidget(self.save_size, 5, 0, 1, 1)

        self.notification = QCheckBox(self.settings_group)
        self.notification.setObjectName(u"notification")

        self.behavior_layout.addWidget(self.notification, 4, 0, 1, 1)

        self.log_games = QCheckBox(self.settings_group)
        self.log_games.setObjectName(u"log_games")

        self.behavior_layout.addWidget(self.log_games, 6, 0, 1, 1)

        self.sys_tray = QCheckBox(self.settings_group)
        self.sys_tray.setObjectName(u"sys_tray")

        self.behavior_layout.addWidget(self.sys_tray, 0, 0, 1, 1)

        self.auto_update = QCheckBox(self.settings_group)
        self.auto_update.setObjectName(u"auto_update")

        self.behavior_layout.addWidget(self.auto_update, 1, 0, 1, 1)

        self.confirm_start = QCheckBox(self.settings_group)
        self.confirm_start.setObjectName(u"confirm_start")

        self.behavior_layout.addWidget(self.confirm_start, 2, 0, 1, 1)

        self.auto_sync_cloud = QCheckBox(self.settings_group)
        self.auto_sync_cloud.setObjectName(u"auto_sync_cloud")

        self.behavior_layout.addWidget(self.auto_sync_cloud, 3, 0, 1, 1)

        self.settings_hspacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.behavior_layout.addItem(self.settings_hspacer, 2, 1, 2, 1)


        self.left_layout.addWidget(self.settings_group)

        self.left_vspacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.left_layout.addItem(self.left_vspacer)


        self.main_layout.addLayout(self.left_layout)

        self.right_layout = QVBoxLayout()
        self.right_layout.setObjectName(u"right_layout")
        self.log_dir_group = QGroupBox(RareSettings)
        self.log_dir_group.setObjectName(u"log_dir_group")
        self.log_dir_layout = QVBoxLayout(self.log_dir_group)
        self.log_dir_layout.setObjectName(u"log_dir_layout")
        self.log_dir_open_button = QPushButton(self.log_dir_group)
        self.log_dir_open_button.setObjectName(u"log_dir_open_button")

        self.log_dir_layout.addWidget(self.log_dir_open_button)

        self.log_dir_clean_button = QPushButton(self.log_dir_group)
        self.log_dir_clean_button.setObjectName(u"log_dir_clean_button")

        self.log_dir_layout.addWidget(self.log_dir_clean_button)

        self.log_dir_size_label = QLabel(self.log_dir_group)
        self.log_dir_size_label.setObjectName(u"log_dir_size_label")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.log_dir_size_label.sizePolicy().hasHeightForWidth())
        self.log_dir_size_label.setSizePolicy(sizePolicy1)
        self.log_dir_size_label.setText(u"")
        self.log_dir_size_label.setWordWrap(True)

        self.log_dir_layout.addWidget(self.log_dir_size_label)


        self.right_layout.addWidget(self.log_dir_group)

        self.groupBox = QGroupBox(RareSettings)
        self.groupBox.setObjectName(u"groupBox")
        self.shortcuts_layout = QVBoxLayout(self.groupBox)
        self.shortcuts_layout.setObjectName(u"shortcuts_layout")
        self.desktop_link_btn = QPushButton(self.groupBox)
        self.desktop_link_btn.setObjectName(u"desktop_link_btn")

        self.shortcuts_layout.addWidget(self.desktop_link_btn)

        self.startmenu_link_btn = QPushButton(self.groupBox)
        self.startmenu_link_btn.setObjectName(u"startmenu_link_btn")

        self.shortcuts_layout.addWidget(self.startmenu_link_btn)


        self.right_layout.addWidget(self.groupBox)

        self.right_vspacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.right_layout.addItem(self.right_vspacer)


        self.main_layout.addLayout(self.right_layout)


        self.retranslateUi(RareSettings)
    # setupUi

    def retranslateUi(self, RareSettings):
        self.interface_group.setTitle(QCoreApplication.translate("RareSettings", u"Interface", None))
        self.lang_label.setText(QCoreApplication.translate("RareSettings", u"Language", None))
        self.color_label.setText(QCoreApplication.translate("RareSettings", u"Color scheme", None))
        self.style_label.setText(QCoreApplication.translate("RareSettings", u"Style sheet", None))
        self.interface_info.setText(QCoreApplication.translate("RareSettings", u"Restart Rare to apply changes.", None))
        self.view_label.setText(QCoreApplication.translate("RareSettings", u"Library view", None))
        self.settings_group.setTitle(QCoreApplication.translate("RareSettings", u"Behavior", None))
        self.save_size.setText(QCoreApplication.translate("RareSettings", u"Restore window size on application startup", None))
        self.notification.setText(QCoreApplication.translate("RareSettings", u"Show notifications when downloads complete", None))
        self.log_games.setText(QCoreApplication.translate("RareSettings", u"Show console window when launching games", None))
        self.sys_tray.setText(QCoreApplication.translate("RareSettings", u"Close to system tray", None))
        self.auto_update.setText(QCoreApplication.translate("RareSettings", u"Queue game updates on application startup", None))
        self.confirm_start.setText(QCoreApplication.translate("RareSettings", u"Confirm before launching games", None))
        self.auto_sync_cloud.setText(QCoreApplication.translate("RareSettings", u"Automatically upload/download cloud saves", None))
        self.log_dir_group.setTitle(QCoreApplication.translate("RareSettings", u"Logs", None))
        self.log_dir_open_button.setText(QCoreApplication.translate("RareSettings", u"Open log folder", None))
        self.log_dir_clean_button.setText(QCoreApplication.translate("RareSettings", u"Clean log folder", None))
        self.groupBox.setTitle(QCoreApplication.translate("RareSettings", u"Shortcuts", None))
        self.desktop_link_btn.setText(QCoreApplication.translate("RareSettings", u"Create on desktop", None))
        self.startmenu_link_btn.setText(QCoreApplication.translate("RareSettings", u"Create in menu", None))
        pass
    # retranslateUi




from enum import Enum
from typing import Tuple

from PySide6.QtCore import QSize


class ImageType(Enum):
    Tall = 0
    Wide = 1
    Icon = 3
    Logo = 2


class ImageSize:
    class Preset:
        def __init__(self, divisor: float, pixel_ratio: float, orientation: ImageType = ImageType.Tall,
                     base: 'ImageSize.Preset' = None):
            self.__divisor = divisor
            self.__pixel_ratio = pixel_ratio
            if orientation == ImageType.Tall:
                self.__img_factor = 67
                self.__size = QSize(self.__img_factor * 3, self.__img_factor * 4) * (pixel_ratio / divisor)
            if orientation == ImageType.Wide:
                self.__img_factor = 34
                self.__size = QSize(self.__img_factor * 16, self.__img_factor * 9) * (pixel_ratio / divisor)
            if orientation == ImageType.Icon:
                self.__img_factor = 128
                self.__size = QSize(self.__img_factor * 1, self.__img_factor * 1) * (pixel_ratio / divisor)
            self.__orientation = orientation
            # lk: for prettier images set this to true
            # self.__smooth_transform: bool = True
            self.__smooth_transform = divisor <= 2
            self.__base = base if base is not None else self

        def __eq__(self, other: 'ImageSize.Preset'):
            return (
                self.__size == other.size
                and self.__divisor == other.divisor
                and self.__smooth_transform == other.smooth
                and self.__pixel_ratio == other.pixel_ratio
            )

        @property
        def size(self) -> QSize:
            return self.__size

        @property
        def divisor(self) -> float:
            return self.__divisor

        @property
        def smooth(self) -> bool:
            return self.__smooth_transform

        @property
        def pixel_ratio(self) -> float:
            return self.__pixel_ratio

        @property
        def orientation(self) -> ImageType:
            return self.__orientation

        @property
        def aspect_ratio(self) -> Tuple[int, int]:
            if self.__orientation == ImageType.Tall:
                return 3, 4
            elif self.__orientation == ImageType.Wide:
                return 16, 9
            else:
                return 0, 0

        @property
        def base(self) -> 'ImageSize.Preset':
            return self.__base

    Tall = Preset(1, 1)
    """! @brief Size and pixel ratio of the image on disk"""

    DisplayTall = Preset(1, 1, base=Tall)
    """! @brief Size and pixel ratio for displaying"""

    LibraryTall = Preset(1.21, 1, base=Tall)
    """! @brief Same as Display"""

    Wide = Preset(1, 1, ImageType.Wide)
    """! @brief Size and pixel ratio for wide 16/9 image on disk"""

    DisplayWide = Preset(2, 1, ImageType.Wide, base=Wide)
    """! @brief Size and pixel ratio for wide 16/9 image display"""

    LibraryWide = Preset(2.41, 1, ImageType.Wide, base=Wide)

    Icon = Preset(1, 1, ImageType.Icon)
    """! @brief Size and pixel ratio of the icon on disk"""

    DisplayIcon = Preset(1, 1, ImageType.Icon, base=Icon)
    """! @brief Size and pixel ratio of the icon on disk"""

    LibraryIcon = Preset(2.2, 1, ImageType.Icon, base=Icon)
    """! @brief Size and pixel ratio of the icon on disk"""



import platform as pf
import re
from logging import getLogger
from typing import Tuple, Set

from PySide6.QtCore import QObject, Signal, QThreadPool, QSettings, Slot
from PySide6.QtGui import QShowEvent, QHideEvent
from PySide6.QtWidgets import QSizePolicy, QWidget, QFileDialog, QMessageBox

from rare.models.options import options
from rare.shared import LegendaryCoreSingleton
from rare.shared.workers.worker import Worker
from rare.ui.components.tabs.settings.legendary import Ui_LegendarySettings
from rare.utils.misc import format_size
from rare.widgets.indicator_edit import PathEdit, IndicatorLineEdit, IndicatorReasonsCommon

logger = getLogger("LegendarySettings")


class RefreshGameMetaWorker(Worker):
    class Signals(QObject):
        finished = Signal()

    def __init__(self, platforms: Set[str], include_unreal: bool):
        super(RefreshGameMetaWorker, self).__init__()
        self.signals = RefreshGameMetaWorker.Signals()
        self.core = LegendaryCoreSingleton()
        self.platforms = platforms if platforms else {"Windows"}
        self.skip_ue = not include_unreal

    def run_real(self) -> None:
        for platform in self.platforms:
            self.core.get_game_and_dlc_list(
                True, platform=platform, force_refresh=True, skip_ue=self.skip_ue
            )
        self.signals.finished.emit()


class LegendarySettings(QWidget):
    def __init__(self, parent=None):
        super(LegendarySettings, self).__init__(parent=parent)
        self.ui = Ui_LegendarySettings()
        self.ui.setupUi(self)
        self.settings = QSettings(self)

        self.core = LegendaryCoreSingleton()

        # Platform specific installation directory for macOS games
        if pf.system() == "Darwin":
            self.mac_install_dir = PathEdit(
                path=self.core.get_default_install_dir("Mac"),
                placeholder=self.tr("Default installation folder for macOS games"),
                file_mode=QFileDialog.FileMode.Directory,
                save_func=self.__mac_path_save,
            )
            self.ui.install_dir_layout.addWidget(self.mac_install_dir)

        # Platform-independent installation directory
        self.install_dir = PathEdit(
            path=self.core.get_default_install_dir(),
            placeholder=self.tr("Default installation folder for Windows games"),
            file_mode=QFileDialog.FileMode.Directory,
            save_func=self.__win_path_save,
        )
        self.ui.install_dir_layout.addWidget(self.install_dir)

        # Max Workers
        self.ui.max_worker_spin.setValue(
            self.core.lgd.config["Legendary"].getint("max_workers", fallback=0)
        )
        self.ui.max_worker_spin.valueChanged.connect(self.max_worker_save)

        # Max memory
        self.ui.max_memory_spin.setValue(
            self.core.lgd.config["Legendary"].getint("max_memory", fallback=0)
        )
        self.ui.max_memory_spin.valueChanged.connect(self.max_memory_save)

        # Preferred CDN
        self.ui.preferred_cdn_line.setText(
            self.core.lgd.config["Legendary"].get("preferred_cdn", fallback="")
        )
        self.ui.preferred_cdn_line.textChanged.connect(self.preferred_cdn_save)

        # Disable HTTPS
        self.ui.disable_https_check.setChecked(
            self.core.lgd.config["Legendary"].getboolean("disable_https", fallback=False)
        )
        self.ui.disable_https_check.stateChanged.connect(self.disable_https_save)

        # Clean metadata
        self.ui.clean_button.clicked.connect(lambda: self.clean_metadata(keep_manifests=False))
        self.ui.clean_keep_manifests_button.clicked.connect(lambda: self.clean_metadata(keep_manifests=True))

        self.locale_edit = IndicatorLineEdit(
            f"{self.core.language_code}-{self.core.country_code}",
            edit_func=self.locale_edit_cb,
            save_func=self.locale_save_cb,
            horiz_policy=QSizePolicy.Policy.Minimum,
            parent=self,
        )
        self.ui.locale_layout.addWidget(self.locale_edit)

        self.ui.fetch_win32_check.setChecked(self.settings.value(*options.win32_meta))
        self.ui.fetch_win32_check.stateChanged.connect(
            lambda: self.settings.setValue(options.win32_meta.key, self.ui.fetch_win32_check.isChecked())
        )

        self.ui.fetch_macos_check.setChecked(self.settings.value(*options.macos_meta))
        self.ui.fetch_macos_check.stateChanged.connect(
            lambda: self.settings.setValue(options.macos_meta.key, self.ui.fetch_macos_check.isChecked())
        )
        self.ui.fetch_macos_check.setDisabled(pf.system() == "Darwin")

        self.ui.fetch_unreal_check.setChecked(self.settings.value(*options.unreal_meta))
        self.ui.fetch_unreal_check.stateChanged.connect(
            lambda: self.settings.setValue(options.unreal_meta.key, self.ui.fetch_unreal_check.isChecked())
        )

        self.ui.exclude_non_asset_check.setChecked(self.settings.value(*options.exclude_non_asset))
        self.ui.exclude_non_asset_check.stateChanged.connect(
            lambda: self.settings.setValue(options.exclude_non_asset.key, self.ui.exclude_non_asset_check.isChecked())
        )

        self.ui.exclude_entitlements_check.setChecked(self.settings.value(*options.exclude_entitlements))
        self.ui.exclude_entitlements_check.stateChanged.connect(
            lambda: self.settings.setValue(options.exclude_entitlements.key, self.ui.exclude_entitlements_check.isChecked())
        )

        self.ui.refresh_metadata_button.clicked.connect(self.refresh_metadata)
        # FIXME: Disable the button for now because it interferes with RareCore
        self.ui.refresh_metadata_button.setEnabled(False)
        self.ui.refresh_metadata_button.setVisible(False)

    def showEvent(self, a0: QShowEvent):
        if a0.spontaneous():
            return super().showEvent(a0)
        return super().showEvent(a0)

    def hideEvent(self, a0: QHideEvent):
        if a0.spontaneous():
            return super().hideEvent(a0)
        self.core.lgd.save_config()
        return super().hideEvent(a0)

    def refresh_metadata(self):
        self.ui.refresh_metadata_button.setDisabled(True)
        platforms = set()
        if self.ui.fetch_win32_check.isChecked():
            platforms.add("Win32")
        if self.ui.fetch_macos_check.isChecked():
            platforms.add("Mac")
        worker = RefreshGameMetaWorker(platforms, self.ui.fetch_unreal_check.isChecked())
        worker.signals.finished.connect(lambda: self.ui.refresh_metadata_button.setDisabled(False))
        QThreadPool.globalInstance().start(worker)

    @staticmethod
    def locale_edit_cb(text: str) -> Tuple[bool, str, int]:
        if text:
            if re.match("^[a-zA-Z]{2,3}[-_][a-zA-Z]{2,3}$", text):
                language, country = text.split("-" if "-" in text else "_")
                text = "-".join([language.lower(), country.upper()])
            if bool(re.match("^[a-z]{2,3}-[A-Z]{2,3}$", text)):
                return True, text, IndicatorReasonsCommon.VALID
            else:
                return False, text, IndicatorReasonsCommon.WRONG_FORMAT
        else:
            return True, text, IndicatorReasonsCommon.VALID

    def locale_save_cb(self, text: str):
        if text:
            self.core.egs.language_code, self.core.egs.country_code = text.split("-")
            self.core.lgd.config.set("Legendary", "locale", text)
        else:
            self.core.lgd.config.remove_option("Legendary", "locale")

    @Slot(str)
    def __mac_path_save(self, text: str) -> None:
        self.__path_save(text, "mac_install_dir")

    @Slot(str)
    def __win_path_save(self, text: str) -> None:
        self.__path_save(text, "install_dir")
        if pf.system() != "Darwin":
            self.__mac_path_save(text)

    def __path_save(self, text: str, option: str):
        if text:
            self.core.lgd.config.set("Legendary", option, text)
        else:
            self.core.lgd.config.remove_option("Legendary", option)

    @Slot(int)
    def max_worker_save(self, workers: int):
        if workers:
            self.core.lgd.config.set("Legendary", "max_workers", str(workers))
        else:
            self.core.lgd.config.remove_option("Legendary", "max_workers")

    @Slot(int)
    def max_memory_save(self, memory: int):
        if memory:
            self.core.lgd.config.set("Legendary", "max_memory", str(memory))
        else:
            self.core.lgd.config.remove_option("Legendary", "max_memory")

    @Slot(str)
    def preferred_cdn_save(self, cdn: str):
        if cdn:
            self.core.lgd.config.set("Legendary", "preferred_cdn", cdn.strip())
        else:
            self.core.lgd.config.remove_option("Legendary", "preferred_cdn")

    @Slot(int)
    def disable_https_save(self, checked: int):
        self.core.lgd.config.set(
            "Legendary", "disable_https", str(bool(checked)).lower()
        )

    def clean_metadata(self, keep_manifests: bool):
        before = self.core.lgd.get_dir_size()
        logger.debug("Removing app metadata...")
        app_names = {g.app_name for g in self.core.get_assets(update_assets=False)}
        self.core.lgd.clean_metadata(app_names)

        if not keep_manifests:
            logger.debug("Removing manifests...")
            installed = [
                (ig.app_name, ig.version, ig.platform) for ig in self.core.get_installed_list()
            ]
            installed.extend(
                (ig.app_name, ig.version, ig.platform) for ig in self.core.get_installed_dlc_list()
            )
            self.core.lgd.clean_manifests(installed)

        logger.debug("Removing tmp data")
        self.core.lgd.clean_tmp_data()

        after = self.core.lgd.get_dir_size()
        logger.info(
            f"Cleanup complete! Removed {(before - after) / 1024 / 1024:.02f} MiB."
        )
        if (before - after) > 0:
            QMessageBox.information(
                self,
                "Cleanup",
                self.tr("Cleanup complete! Successfully removed {}").format(
                    format_size(before - after)
                ),
            )
        else:
            QMessageBox.information(self, "Cleanup", "Nothing to clean")



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'import_group.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QFormLayout,
    QGroupBox, QHBoxLayout, QLabel, QPushButton,
    QSizePolicy, QWidget)

class Ui_ImportGroup(object):
    def setupUi(self, ImportGroup):
        if not ImportGroup.objectName():
            ImportGroup.setObjectName(u"ImportGroup")
        ImportGroup.resize(651, 232)
        ImportGroup.setWindowTitle(u"ImportGroup")
        ImportGroup.setWindowFilePath(u"")
        self.import_layout = QFormLayout(ImportGroup)
        self.import_layout.setObjectName(u"import_layout")
        self.import_layout.setLabelAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)
        self.path_edit_label = QLabel(ImportGroup)
        self.path_edit_label.setObjectName(u"path_edit_label")

        self.import_layout.setWidget(0, QFormLayout.LabelRole, self.path_edit_label)

        self.app_name_label = QLabel(ImportGroup)
        self.app_name_label.setObjectName(u"app_name_label")

        self.import_layout.setWidget(1, QFormLayout.LabelRole, self.app_name_label)

        self.import_folder_label = QLabel(ImportGroup)
        self.import_folder_label.setObjectName(u"import_folder_label")

        self.import_layout.setWidget(3, QFormLayout.LabelRole, self.import_folder_label)

        self.import_folder_check = QCheckBox(ImportGroup)
        self.import_folder_check.setObjectName(u"import_folder_check")
        font = QFont()
        font.setItalic(True)
        self.import_folder_check.setFont(font)

        self.import_layout.setWidget(3, QFormLayout.ItemRole.FieldRole, self.import_folder_check)

        self.import_dlcs_label = QLabel(ImportGroup)
        self.import_dlcs_label.setObjectName(u"import_dlcs_label")

        self.import_layout.setWidget(4, QFormLayout.LabelRole, self.import_dlcs_label)

        self.import_dlcs_check = QCheckBox(ImportGroup)
        self.import_dlcs_check.setObjectName(u"import_dlcs_check")
        self.import_dlcs_check.setFont(font)

        self.import_layout.setWidget(4, QFormLayout.ItemRole.FieldRole, self.import_dlcs_check)

        self.import_force_label = QLabel(ImportGroup)
        self.import_force_label.setObjectName(u"import_force_label")

        self.import_layout.setWidget(5, QFormLayout.LabelRole, self.import_force_label)

        self.import_force_check = QCheckBox(ImportGroup)
        self.import_force_check.setObjectName(u"import_force_check")
        self.import_force_check.setFont(font)

        self.import_layout.setWidget(5, QFormLayout.ItemRole.FieldRole, self.import_force_check)

        self.import_button_label = QLabel(ImportGroup)
        self.import_button_label.setObjectName(u"import_button_label")
        self.import_button_label.setText(u"error")

        self.import_layout.setWidget(6, QFormLayout.LabelRole, self.import_button_label)

        self.button_info_layout = QHBoxLayout()
        self.button_info_layout.setObjectName(u"button_info_layout")
        self.import_button = QPushButton(ImportGroup)
        self.import_button.setObjectName(u"import_button")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.import_button.sizePolicy().hasHeightForWidth())
        self.import_button.setSizePolicy(sizePolicy)

        self.button_info_layout.addWidget(self.import_button)


        self.import_layout.setLayout(6, QFormLayout.ItemRole.FieldRole, self.button_info_layout)

        self.platform_label = QLabel(ImportGroup)
        self.platform_label.setObjectName(u"platform_label")

        self.import_layout.setWidget(2, QFormLayout.LabelRole, self.platform_label)

        self.platform_layout = QHBoxLayout()
        self.platform_layout.setObjectName(u"platform_layout")
        self.platform_combo = QComboBox(ImportGroup)
        self.platform_combo.setObjectName(u"platform_combo")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.platform_combo.sizePolicy().hasHeightForWidth())
        self.platform_combo.setSizePolicy(sizePolicy1)

        self.platform_layout.addWidget(self.platform_combo)

        self.platform_tooltip = QLabel(ImportGroup)
        self.platform_tooltip.setObjectName(u"platform_tooltip")
        self.platform_tooltip.setFont(font)

        self.platform_layout.addWidget(self.platform_tooltip)


        self.import_layout.setLayout(2, QFormLayout.ItemRole.FieldRole, self.platform_layout)


        self.retranslateUi(ImportGroup)
    # setupUi

    def retranslateUi(self, ImportGroup):
        ImportGroup.setTitle(QCoreApplication.translate("ImportGroup", u"Import EGL game from a directory", None))
        self.path_edit_label.setText(QCoreApplication.translate("ImportGroup", u"Installation path", None))
        self.app_name_label.setText(QCoreApplication.translate("ImportGroup", u"Override app name", None))
        self.import_folder_label.setText(QCoreApplication.translate("ImportGroup", u"Import all games", None))
        self.import_folder_check.setText(QCoreApplication.translate("ImportGroup", u"Scan the installation path for game folders and import them", None))
        self.import_dlcs_label.setText(QCoreApplication.translate("ImportGroup", u"Import DLCs", None))
        self.import_dlcs_check.setText(QCoreApplication.translate("ImportGroup", u"If a game has DLCs, try to import them too", None))
        self.import_force_label.setText(QCoreApplication.translate("ImportGroup", u"Force import", None))
        self.import_force_check.setText(QCoreApplication.translate("ImportGroup", u"Import game despite missing files", None))
        self.import_button.setText(QCoreApplication.translate("ImportGroup", u"Import game(s)", None))
        self.platform_label.setText(QCoreApplication.translate("ImportGroup", u"Platform", None))
        self.platform_tooltip.setText(QCoreApplication.translate("ImportGroup", u"Select the native platform of the game", None))
    # retranslateUi




import os
import platform as pf
import subprocess
from configparser import ConfigParser
from logging import getLogger
from typing import Mapping, Dict, List, Tuple

from PySide6.QtCore import QProcess, QProcessEnvironment

from rare.utils import config_helper as config
if pf.system() != "Windows":
    from . import wine
    if pf.system() in {"Linux", "FreeBSD"}:
        from . import steam

logger = getLogger("CompatUtils")


# this is a copied function from legendary.utils.wine_helpers, but registry file can be specified
def read_registry(registry: str, prefix: str) -> ConfigParser:
    accepted = ["system.reg", "user.reg"]
    if registry not in accepted:
        raise RuntimeError(f'Unknown target "{registry}" not in {accepted}')
    reg = ConfigParser(comment_prefixes=(';', '#', '/', 'WINE'), allow_no_value=True,
                       strict=False)
    reg.optionxform = str
    reg.read(os.path.join(prefix, 'system.reg'))
    return reg


def get_configured_qprocess(command: List[str], environment: Mapping) -> QProcess:
    logger.debug("Executing command: %s", command)
    proc = QProcess()
    proc.setProcessChannelMode(QProcess.SeparateChannels)
    penv = QProcessEnvironment()
    for ek, ev in environment.items():
        penv.insert(ek, ev)
    proc.setProcessEnvironment(penv)
    proc.setProgram(command[0])
    proc.setArguments(command[1:])
    return proc


def get_configured_subprocess(command: List[str], environment: Mapping) -> subprocess.Popen:
    logger.debug("Executing command: %s", command)
    return subprocess.Popen(
        command,
        stdin=None,
        stdout=subprocess.PIPE,
        stderr=None,
        env=environment,
        shell=False,
        text=False,
    )


def execute_subprocess(command: List[str], arguments: List[str], environment: Mapping) -> Tuple[str, str]:
    proc = get_configured_subprocess(command + arguments, environment)
    print(proc.args)
    out, err = proc.communicate()
    out, err = out.decode("utf-8", "ignore") if out else "", err.decode("utf-8", "ignore") if err else ""

    # lk: the following is a work-around for wineserver sometimes hanging around after
    proc = get_configured_subprocess(command + ["wineboot", "-e"], environment)
    _, _ = proc.communicate()
    return out, err


def execute_qprocess(command: List[str], arguments: List[str], environment: Mapping) -> Tuple[str, str]:
    proc = get_configured_qprocess(command + arguments, environment)
    proc.start()
    proc.waitForFinished(-1)
    out, err = (
        proc.readAllStandardOutput().data().decode("utf-8", "ignore"),
        proc.readAllStandardError().data().decode("utf-8", "ignore")
    )
    proc.deleteLater()

    # lk: the following is a work-around for wineserver sometimes hanging around after
    proc = get_configured_qprocess(command + ["wineboot", "-e"], environment)
    proc.start()
    proc.waitForFinished(-1)
    proc.deleteLater()

    return out, err


def execute(command: List[str], arguments: List[str], environment: Mapping) -> Tuple[str, str]:
    # Use the current environment if we are in flatpak or our own if we are on host
    # In flatpak our environment is passed through `flatpak-spawn` arguments
    if os.environ.get("container") == "flatpak":
        flatpak_command = ["flatpak-spawn", "--host"]
        flatpak_command.extend(f"--env={name}={value}" for name, value in environment.items())
        _command = flatpak_command + command
        _environment = os.environ.copy()
    else:
        _command = command
        _environment = environment

    try:
        out, err = execute_qprocess(_command, arguments, _environment)
    except Exception as e:
        out, err = "", str(e)

    return out, err


def resolve_path(command: List[str], environment: Mapping, path: str) -> str:
    path = path.strip().replace("/", "\\")
    # lk: if path does not exist form
    arguments = ["cmd.exe", "/c", "echo", path]
    # lk: if path exists and needs a case-sensitive interpretation form
    # cmd = [wine_cmd, 'cmd', '/c', f'cd {path} & cd']
    out, err = execute(command, arguments, environment)
    out, err = out.strip(), err.strip()
    if not out:
        logger.error("Failed to resolve wine path due to \"%s\"", err)
        return out
    return out.strip('"')


def query_reg_path(wine_exec: str, wine_env: Mapping, reg_path: str):
    raise NotImplementedError


def query_reg_key(command: List[str], environment: Mapping, reg_path: str, reg_key) -> str:
    arguments = ["reg.exe", "query", reg_path, "/v", reg_key]
    out, err = execute(command, arguments, environment)
    out, err = out.strip(), err.strip()
    if not out:
        logger.error("Failed to query registry key due to \"%s\"", err)
        return out
    lines = out.split("\n")
    keys: Dict = {}
    for line in lines:
        if line.startswith(" "*4):
            key = [x for x in line.split(" "*4, 3) if bool(x)]
            keys.update({key[0]: key[2]})
    return keys.get(reg_key, "")


def convert_to_windows_path(wine_exec: str, wine_env: Mapping, path: str) -> str:
    raise NotImplementedError


def convert_to_unix_path(command: List[str], environment: Mapping, path: str) -> str:
    path = path.strip().strip('"')
    arguments = ["winepath.exe", "-u", path]
    out, err = execute(command, arguments, environment)
    out, err = out.strip(), err.strip()
    if not out:
        logger.error("Failed to convert to unix path due to \"%s\"", err)
    return os.path.realpath(out) if (out := out.strip()) else out


def get_host_environment(app_environment: Dict, silent: bool = True) -> Dict:
    # Get a clean environment if we are in flatpak, this environment will be passed
    # to `flatpak-spawn`, otherwise use the system's.
    _environ = {} if os.environ.get("container") == "flatpak" else os.environ.copy()
    _environ.update(app_environment)
    if silent:
        _environ["WINEESYNC"] = "0"
        _environ["WINEFSYNC"] = "0"
        _environ["WINE_DISABLE_FAST_SYNC"] = "1"
        _environ["WINEDEBUG"] = "-all"
        _environ["WINEDLLOVERRIDES"] = "winemenubuilder=d;mscoree=d;mshtml=d;"
        # lk: pressure-vessel complains about this but it doesn't fail due to it
        #_environ["DISPLAY"] = ""
    return _environ



from abc import abstractmethod
from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum
from logging import getLogger
from typing import Optional, List, Tuple

from PySide6.QtCore import QObject, Signal, QRunnable, QThreadPool, QSettings
from legendary.lfs import eos
from legendary.models.game import SaveGameFile, SaveGameStatus, Game, InstalledGame
from legendary.utils.selective_dl import get_sdl_appname

from rare.models.options import options
from rare.lgndr.core import LegendaryCore
from rare.models.install import UninstallOptionsModel, InstallOptionsModel

logger = getLogger("RareGameBase")


@dataclass
class RareSaveGame:
    file: Optional[SaveGameFile]
    status: SaveGameStatus = SaveGameStatus.NO_SAVE
    dt_local: Optional[datetime] = None
    dt_remote: Optional[datetime] = None
    description: Optional[str] = ""


class RareGameBase(QObject):

    class State(IntEnum):
        IDLE = 0
        RUNNING = 1
        DOWNLOADING = 2
        VERIFYING = 3
        MOVING = 4
        UNINSTALLING = 5
        SYNCING = 6

    class Signals(QObject):
        class Progress(QObject):
            start = Signal()
            update = Signal(int)
            finish = Signal(bool)

        class Widget(QObject):
            update = Signal()

        class Download(QObject):
            enqueue = Signal(str)
            dequeue = Signal(str)

        class Game(QObject):
            install = Signal(InstallOptionsModel)
            installed = Signal(str)
            uninstall = Signal(UninstallOptionsModel)
            uninstalled = Signal(str)
            launched = Signal(str)
            finished = Signal(str)

        def __init__(self, parent=None):
            super(RareGameBase.Signals, self).__init__(parent=parent)
            self.progress = RareGameBase.Signals.Progress(self)
            self.widget = RareGameBase.Signals.Widget(self)
            self.download = RareGameBase.Signals.Download(self)
            self.game = RareGameBase.Signals.Game(self)

    def __init__(self, legendary_core: LegendaryCore, game: Game, parent=None):
        super(RareGameBase, self).__init__(parent=parent)
        self.signals = RareGameBase.Signals(self)
        self.core = legendary_core
        self.game: Game = game
        self.igame: InstalledGame = None
        self._state = RareGameBase.State.IDLE

    @property
    def state(self) -> 'RareGameBase.State':
        return self._state

    @state.setter
    def state(self, state: 'RareGameBase.State'):
        if state != self._state:
            self._state = state
            self.signals.widget.update.emit()

    @property
    def is_idle(self):
        return self.state == RareGameBase.State.IDLE

    @property
    def app_name(self) -> str:
        return self.game.app_name

    @property
    def app_title(self) -> str:
        return self.game.app_title

    @property
    @abstractmethod
    def is_installed(self) -> bool:
        pass

    @abstractmethod
    def set_installed(self, installed: bool) -> None:
        pass

    @property
    def platforms(self) -> Tuple:
        """!
        @brief Property that holds the platforms a game is available for

        @return Tuple
        """
        return tuple(self.game.asset_infos.keys())

    @property
    def default_platform(self) -> str:
        return self.core.default_platform if self.core.default_platform in self.platforms else "Windows"

    @property
    def is_mac(self) -> bool:
        """!
        @brief Property to report if Game has a mac version

        @return bool
        """
        return "Mac" in self.game.asset_infos.keys()

    @property
    def is_win32(self) -> bool:
        """!
        @brief Property to report if Game is 32bit game

        @return bool
        """
        return "Win32" in self.game.asset_infos.keys()

    @property
    def is_origin(self) -> bool:
        """!
        @brief Property to report if a Game is an Origin game

        Legendary and by extenstion Rare can't launch Origin games directly,
        it just launches the Origin client and thus requires a bit of a special
        handling to let the user know.

        @return bool If the game is an Origin game
        """
        return self.game.third_party_store in {"Origin", "the EA app"}

    @property
    def is_overlay(self):
        return self.app_name == eos.EOSOverlayApp.app_name

    @property
    def is_dlc(self) -> bool:
        """!
        @brief Property to report if Game is a dlc

        @return bool
        """
        return self.game.is_dlc

    @property
    def is_launchable_addon(self) -> bool:
        # lk: the attribute doesn't exist in the currently released version
        # FIXME: remove after legendary >= 0.20.35
        try:
            return self.game.is_launchable_addon
        except AttributeError:
            return False

    @property
    def sdl_name(self) -> Optional[str]:
        return get_sdl_appname(self.app_name)

    @property
    def version(self) -> str:
        """!
        @brief Reports the currently installed version of the Game

        If InstalledGame reports the currently installed version, which might be
        different from the remote version available from EGS. For not installed Games
        it reports the already known version.

        @return str The current version of the game
        """
        return self.igame.version if self.igame is not None else self.game.app_version(self.default_platform)

    @property
    def install_path(self) -> Optional[str]:
        if self.igame:
            return self.igame.install_path
        return None


class RareGameSlim(RareGameBase):

    def __init__(self, legendary_core: LegendaryCore, game: Game, parent=None):
        super(RareGameSlim, self).__init__(legendary_core, game, parent=parent)
        # None if origin or not installed
        self.igame: Optional[InstalledGame] = self.core.get_installed_game(game.app_name)
        self.saves: List[RareSaveGame] = []

    @property
    def is_installed(self) -> bool:
        if self.is_origin:
            return True
        return self.igame is not None

    def set_installed(self, installed: bool) -> None:
        pass

    @property
    def auto_sync_saves(self):
        auto_sync_cloud = QSettings(self).value(
            f"{self.app_name}/{options.auto_sync_cloud.key}",
            QSettings(self).value(*options.auto_sync_cloud),
            options.auto_sync_cloud.dtype
        )
        return self.supports_cloud_saves and auto_sync_cloud

    @property
    def save_path(self) -> Optional[str]:
        if self.igame is not None:
            return self.igame.save_path
        return None

    @property
    def latest_save(self) -> RareSaveGame:
        if self.saves:
            saves = sorted(self.saves, key=lambda s: s.file.datetime, reverse=True)
            return saves[0]
        return RareSaveGame(None)

    @property
    def save_game_state(self) -> Tuple[SaveGameStatus, Tuple[Optional[datetime], Optional[datetime]]]:
        if self.save_path:
            latest = self.latest_save
            # lk: if the save path wasn't known at startup, dt_local will be None
            # In that case resolve the save again before returning
            latest.status, (latest.dt_local, latest.dt_remote) = self.core.check_savegame_state(
                self.save_path, latest.file
            )
            return latest.status, (latest.dt_local, latest.dt_remote)
        return SaveGameStatus.NO_SAVE, (None, None)

    def upload_saves(self, thread=True):
        if not self.supports_cloud_saves:
            return
        if self.state == RareGameSlim.State.SYNCING:
            logger.error(f"{self.app_title} is already syncing")
            return

        status, (dt_local, dt_remote) = self.save_game_state
        if status == SaveGameStatus.NO_SAVE or not dt_local:
            logger.warning("Can't upload non existing save")
            return

        def _upload():
            logger.info(f"Uploading save for {self.app_title}")
            self.state = RareGameSlim.State.SYNCING
            self.core.upload_save(self.app_name, self.igame.save_path, dt_local)
            self.state = RareGameSlim.State.IDLE
            self.update_saves()

        if thread:
            worker = QRunnable.create(_upload)
            QThreadPool.globalInstance().start(worker)
        else:
            _upload()

    def download_saves(self, thread=True):
        if not self.supports_cloud_saves:
            return
        if self.state == RareGameSlim.State.SYNCING:
            logger.error(f"{self.app_title} is already syncing")
            return

        status, (dt_local, dt_remote) = self.save_game_state
        if status == SaveGameStatus.NO_SAVE or not dt_remote:
            logger.error("Can't download non existing save")
            return

        def _download():
            logger.info(f"Downloading save for {self.app_title}")
            self.state = RareGameSlim.State.SYNCING
            self.core.download_saves(self.app_name, self.latest_save.file.manifest_name, self.save_path)
            self.state = RareGameSlim.State.IDLE
            self.update_saves()

        if thread:
            worker = QRunnable.create(_download)
            QThreadPool.globalInstance().start(worker)
        else:
            _download()

    def load_saves(self, saves: List[SaveGameFile]):
        """ Use only in a thread """
        self.saves.clear()
        for save in saves:
            if self.save_path:
                status, (dt_local, dt_remote) = self.core.check_savegame_state(self.save_path, save)
                rsave = RareSaveGame(save, status, dt_local, dt_remote)
            else:
                rsave = RareSaveGame(save, SaveGameStatus.SAME_AGE, dt_local=None, dt_remote=save.datetime)
            self.saves.append(rsave)
        self.signals.widget.update.emit()

    def update_saves(self):
        """ Use only in a thread """
        saves = self.core.get_save_games(self.app_name)
        self.load_saves(saves)

    @property
    def is_save_up_to_date(self):
        if not self.saves:
            return True
        status, (_, _) = self.save_game_state
        return status in {SaveGameStatus.SAME_AGE, SaveGameStatus.NO_SAVE}

    @property
    def raw_save_path(self) -> str:
        if self.game.supports_cloud_saves:
            return self.game.metadata.get("customAttributes", {}).get("CloudSaveFolder", {}).get("value")
        return ""

    @property
    def raw_save_path_mac(self) -> str:
        if self.game.supports_mac_cloud_saves:
            return self.game.metadata.get("customAttributes", {}).get('CloudSaveFolder_MAC', {}).get('value')
        return ""

    @property
    def supports_cloud_saves(self):
        return self.game.supports_cloud_saves or self.game.supports_mac_cloud_saves



import logging
from typing import List, Dict

from PySide6.QtCore import Qt, QUrl, Signal
from PySide6.QtGui import QFont, QDesktopServices, QKeyEvent
from PySide6.QtWidgets import (
    QWidget,
    QLabel,
    QPushButton,
    QGridLayout,
    QSizePolicy,
)

from rare.components.tabs.store.api.models.diesel import DieselProduct, DieselProductDetail, DieselSystemDetail
from rare.components.tabs.store.api.models.response import CatalogOfferModel
from rare.components.tabs.store.store_api import StoreAPI
from rare.models.image import ImageSize
from rare.ui.components.tabs.store.details import Ui_StoreDetailsWidget
from rare.utils.misc import qta_icon
from rare.widgets.elide_label import ElideLabel
from rare.widgets.side_tab import SideTabWidget, SideTabContents
from .image import LoadingImageWidget

logger = logging.getLogger("StoreDetails")


class StoreDetailsWidget(QWidget, SideTabContents):
    back_clicked: Signal = Signal()

    # TODO Design
    def __init__(self, installed: List, store_api: StoreAPI, parent=None):
        super(StoreDetailsWidget, self).__init__(parent=parent)
        self.implements_scrollarea = True

        self.ui = Ui_StoreDetailsWidget()
        self.ui.setupUi(self)
        self.ui.main_layout.setContentsMargins(0, 0, 3, 0)

        self.store_api = store_api
        self.installed = installed
        self.catalog_offer: CatalogOfferModel = None

        self.image = LoadingImageWidget(store_api.cached_manager, self)
        self.image.setFixedSize(ImageSize.DisplayTall)
        self.ui.left_layout.insertWidget(0, self.image, alignment=Qt.AlignmentFlag.AlignTop)
        self.ui.left_layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        self.ui.wishlist_button.clicked.connect(self.add_to_wishlist)
        self.ui.store_button.clicked.connect(self.button_clicked)
        self.ui.wishlist_button.setVisible(True)
        self.in_wishlist = False
        self.wishlist = []

        self.requirements_tabs = SideTabWidget(parent=self.ui.requirements_frame)
        self.requirements_tabs.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self.ui.requirements_layout.addWidget(self.requirements_tabs)

        self.ui.back_button.setIcon(qta_icon("fa.chevron-left"))
        self.ui.back_button.clicked.connect(self.back_clicked)

        self.setDisabled(False)

    def handle_wishlist_update(self, wishlist: List[CatalogOfferModel]):
        if wishlist and wishlist[0] == "error":
            return
        self.wishlist = [game.id for game in wishlist]
        if self.id_str in self.wishlist:
            self.in_wishlist = True
            self.ui.wishlist_button.setText(self.tr("Remove from Wishlist"))
        else:
            self.in_wishlist = False

    def update_game(self, offer: CatalogOfferModel):
        self.ui.title.setText(offer.title)
        self.title_str = offer.title
        self.id_str = offer.id
        self.store_api.get_wishlist(self.handle_wishlist_update)

        # lk: delete tabs in reverse order because indices are updated on deletion
        while self.requirements_tabs.count():
            self.requirements_tabs.widget(0).deleteLater()
            self.requirements_tabs.removeTab(0)
        self.requirements_tabs.clear()

        slug = offer.productSlug
        if not slug:
            for mapping in offer.offerMappings:
                if mapping["pageType"] == "productHome":
                    slug = mapping["pageSlug"]
                    break
            else:
                logger.error("Could not get page information")
                slug = ""
        if "/home" in slug:
            slug = slug.replace("/home", "")
        self.slug = slug

        if offer.namespace in self.installed:
            self.ui.store_button.setText(self.tr("Show Game on Epic Page"))
            self.ui.status.setVisible(True)
        else:
            self.ui.store_button.setText(self.tr("Buy Game in Epic Games Store"))
            self.ui.status.setVisible(False)

        self.ui.original_price.setText(self.tr("Loading"))
        # self.title.setText(self.tr("Loading"))
        # self.image.setPixmap(QPixmap())
        is_bundle = False
        for i in offer.categories:
            if "bundles" in i.get("path", ""):
                is_bundle = True

        # init API request
        if slug:
            self.store_api.get_game_config_cms(offer.productSlug, is_bundle, self.data_received)
        # else:
        #     self.data_received({})
        self.catalog_offer = offer

    def add_to_wishlist(self):
        if not self.in_wishlist:
            self.store_api.add_to_wishlist(
                self.catalog_offer.namespace,
                self.catalog_offer.id,
                lambda success: self.ui.wishlist_button.setText(self.tr("Remove from wishlist"))
                if success
                else self.ui.wishlist_button.setText("Something went wrong")
            )
        else:
            self.store_api.remove_from_wishlist(
                self.catalog_offer.namespace,
                self.catalog_offer.id,
                lambda success: self.ui.wishlist_button.setText(self.tr("Add to wishlist"))
                if success
                else self.ui.wishlist_button.setText("Something went wrong"),
            )

    def data_received(self, product: DieselProduct):
        try:
            if product.pages:
                product_data: DieselProductDetail = product.pages[0].data
            else:
                product_data: DieselProductDetail = product.data
        except Exception as e:
            raise e
            logger.error(str(e))

        self.ui.original_price.setFont(self.font())
        price = self.catalog_offer.price.totalPrice.fmtPrice["originalPrice"]
        discount_price = self.catalog_offer.price.totalPrice.fmtPrice["discountPrice"]
        if price == "0" or price == 0:
            self.ui.original_price.setText(self.tr("Free"))
        else:
            self.ui.original_price.setText(price)
        if price != discount_price:
            font = self.font()
            font.setStrikeOut(True)
            self.ui.original_price.setFont(font)
            self.ui.discount_price.setText(
                discount_price
                if discount_price != "0"
                else self.tr("Free")
            )
            self.ui.discount_price.setVisible(True)
        else:
            self.ui.discount_price.setVisible(False)

        requirements = product_data.requirements
        if requirements and requirements.systems:
            for system in requirements.systems:
                req_widget = RequirementsWidget(system, self.requirements_tabs)
                self.requirements_tabs.addTab(req_widget, system.systemType)
            self.ui.requirements_frame.setVisible(True)
        else:
            self.ui.requirements_frame.setVisible(False)

        key_images = self.catalog_offer.keyImages
        img_url = key_images.for_dimensions(self.image.size().width(), self.image.size().height())
        # FIXME: check why there was no tall image
        if img_url:
            self.image.fetchPixmap(img_url.url)

        # self.image_stack.setCurrentIndex(0)
        about = product_data.about
        self.ui.description_label.setMarkdown(about.desciption)
        self.ui.developer.setText(about.developerAttribution)
        # try:
        #     if isinstance(aboudeveloper, list):
        #         self.ui.dev.setText(", ".join(self.game.developer))
        #     else:
        #         self.ui.dev.setText(self.game.developer)
        # except KeyError:
        #     pass
        tags = product_data.unmapped["meta"].get("tags", [])
        self.ui.tags.setText(", ".join(tags))

        # clear Layout
        for b in self.ui.social_links.findChildren(SocialButton, options=Qt.FindChildOption.FindDirectChildrenOnly):
            self.ui.social_links_layout.removeWidget(b)
            b.deleteLater()

        links = product_data.socialLinks
        link_count = 0
        for name, url in links.items():
            if name == "_type":
                continue
            name = name.replace("link", "").lower()
            if name == "homepage":
                icn = qta_icon("mdi.web", "fa.search", scale_factor=1.5)
            else:
                try:
                    icn = qta_icon(f"mdi.{name}", f"fa.{name}", scale_factor=1.5)
                except Exception as e:
                    logger.error(str(e))
                    continue

            button = SocialButton(icn, url, parent=self.ui.social_links)
            self.ui.social_links_layout.addWidget(button)
            link_count += 1

        self.ui.social_links.setEnabled(bool(link_count))

        self.setEnabled(True)

    # def add_wishlist_items(self, wishlist: List[CatalogGameModel]):
    #     wishlist = wishlist["data"]["Wishlist"]["wishlistItems"]["elements"]
    #     for game in wishlist:
    #         self.wishlist.append(game["offer"]["title"])

    def button_clicked(self):
        QDesktopServices.openUrl(QUrl(f"https://www.epicgames.com/store/{self.store_api.language_code}/p/{self.slug}"))

    def keyPressEvent(self, a0: QKeyEvent):
        if a0.key() == Qt.Key.Key_Escape:
            self.back_clicked.emit()


class SocialButton(QPushButton):
    def __init__(self, icn, url, parent=None):
        super(SocialButton, self).__init__(icn, "", parent=parent)
        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.url = url
        self.clicked.connect(lambda: QDesktopServices.openUrl(QUrl(url)))
        self.setToolTip(url)


class RequirementsWidget(QWidget, SideTabContents):
    def __init__(self, system: DieselSystemDetail, parent=None):
        super().__init__(parent=parent)
        self.implements_scrollarea = True

        bold_font = self.font()
        bold_font.setBold(True)

        req_layout = QGridLayout(self)
        min_label = QLabel(self.tr("Minimum"), parent=self)
        min_label.setFont(bold_font)
        rec_label = QLabel(self.tr("Recommend"), parent=self)
        rec_label.setFont(bold_font)
        req_layout.addWidget(min_label, 0, 1)
        req_layout.addWidget(rec_label, 0, 2)
        req_layout.setColumnStretch(1, 2)
        req_layout.setColumnStretch(2, 2)
        for i, detail in enumerate(system.details):
            req_layout.addWidget(QLabel(detail.title, parent=self), i + 1, 0)
            min_label = ElideLabel(detail.minimum, parent=self)
            req_layout.addWidget(min_label, i + 1, 1)
            rec_label = ElideLabel(detail.recommended, parent=self)
            req_layout.addWidget(rec_label, i + 1, 2)
        req_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)



from PySide6.QtCore import QObject, Signal

from .install import InstallOptionsModel, UninstallOptionsModel


class GlobalSignals:

    # set_main_tab_index = Signal(int)  # tab index
    # update_gamelist = Signal(list)

    class ApplicationSignals(QObject):
        # int: exit code
        quit = Signal(int)
        # str: title, str: body
        notify = Signal(str, str)
        # none
        update_tray = Signal()
        # none
        update_statusbar = Signal()
        # str: locale
        # change_translation = Signal(str)

    class GameSignals(QObject):
        # model
        install = Signal(InstallOptionsModel)
        # str: app_name
        installed = Signal(str)
        # model
        uninstall = Signal(UninstallOptionsModel)
        # str: app_name
        uninstalled = Signal(str)

    class DownloadSignals(QObject):
        # str: app_name
        enqueue = Signal(str)
        # str: app_name
        dequeue = Signal(str)

    class DiscordRPCSignals(QObject):
        # str: app_name
        update_presence = Signal(str)
        # str: app_name
        remove_presence = Signal(str)
        # none
        update_settings = Signal()

    def __init__(self):
        self.application = GlobalSignals.ApplicationSignals()
        self.game = GlobalSignals.GameSignals()
        self.download = GlobalSignals.DownloadSignals()
        self.discord_rpc = GlobalSignals.DiscordRPCSignals()

    def deleteLater(self):
        self.application.deleteLater()
        del self.application
        self.game.deleteLater()
        del self.game
        self.download.deleteLater()
        del self.download
        self.discord_rpc.deleteLater()
        del self.discord_rpc



import random
import sys

from setuptools_scm import ScmVersion, get_version

def mknumeric(ver: ScmVersion) -> str:
    return f"{ver.tag}.{ver.distance}"


def mkname() -> str:
    from legendary.core import LegendaryCore

    core = LegendaryCore()
    core.login()

    return " ".join(
        map(
            lambda game: game.app_name,
            random.choices(list(filter(lambda x: len(x.app_name) != 32, core.get_game_list(False))), k=2),
        )
    )


if __name__ == "__main__":
    version = get_version(
        root='..',
        relative_to=__file__,
        git_describe_command=["git", "describe", "--dirty", "--long"],
        version_scheme=mknumeric,
        local_scheme="no-local-version"
    )
    sys.stdout.write(version)
    sys.stdout.write("\n")
    sys.stdout.write(mkname())
    sys.stdout.write("\n")
    sys.exit(0)



import os
from logging import getLogger
from typing import Optional

from PySide6.QtCore import Signal, Qt, QSignalBlocker
from PySide6.QtGui import QShowEvent
from PySide6.QtWidgets import QFileDialog, QFormLayout, QGroupBox

from rare.shared import LegendaryCoreSingleton, GlobalSignalsSingleton
from rare.utils import config_helper as config
from rare.widgets.indicator_edit import PathEdit, IndicatorReasonsCommon

logger = getLogger("WineSettings")


class WineSettings(QGroupBox):
    # str: option key
    environ_changed = Signal(str)

    def __init__(self, parent=None):
        super(WineSettings, self).__init__(parent=parent)
        self.setTitle(self.tr("Wine settings"))

        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()

        self.app_name: Optional[str] = "default"

        # Wine prefix
        self.wine_prefix = PathEdit(
            path="",
            file_mode=QFileDialog.FileMode.Directory,
            edit_func=lambda path: (os.path.isdir(path) or not path, path, IndicatorReasonsCommon.DIR_NOT_EXISTS),
            save_func=self.save_prefix,
        )

        # Wine executable
        self.wine_exec = PathEdit(
            path="",
            file_mode=QFileDialog.FileMode.ExistingFile,
            name_filters=("wine", "wine64"),
            edit_func=lambda text: (os.path.isfile(text) or not text, text, IndicatorReasonsCommon.FILE_NOT_EXISTS),
            save_func=self.save_exec,
        )

        layout = QFormLayout(self)
        layout.addRow(self.tr("Executable"), self.wine_exec)
        layout.addRow(self.tr("Prefix"), self.wine_prefix)
        layout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        layout.setLabelAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        layout.setFormAlignment(Qt.AlignmentFlag.AlignLeading | Qt.AlignmentFlag.AlignTop)

    def showEvent(self, a0: QShowEvent):
        if a0.spontaneous():
            return super().showEvent(a0)

        _ = QSignalBlocker(self.wine_prefix)
        self.wine_prefix.setText(self.load_prefix())
        _ = QSignalBlocker(self.wine_exec)
        self.wine_exec.setText(self.load_exec())
        self.setDisabled(config.get_boolean(self.app_name, "no_wine", fallback=False))

        return super().showEvent(a0)

    def tool_enabled(self, enabled: bool):
        if enabled:
            config.set_boolean(self.app_name, "no_wine", True)
        else:
            config.remove_option(self.app_name, "no_wine")
        self.setDisabled(enabled)

    def load_prefix(self) -> str:
        if self.app_name is None:
            raise RuntimeError
        return config.get_wine_prefix(self.app_name, "")

    def save_prefix(self, path: str) -> None:
        if self.app_name is None:
            raise RuntimeError
        config.save_wine_prefix(self.app_name, path)
        self.environ_changed.emit("WINEPREFIX")

    def load_exec(self) -> str:
        if self.app_name is None:
            raise RuntimeError
        return config.get_option(self.app_name, "wine_executable", "")

    def save_exec(self, text: str) -> None:
        if self.app_name is None:
            raise RuntimeError
        config.save_option(self.app_name, "wine_executable", text)



import logging

from PySide6.QtCore import QSettings, Signal, Slot, QSize, Qt
from PySide6.QtGui import QAction
from PySide6.QtWidgets import (
    QHBoxLayout,
    QWidget,
    QPushButton,
    QLabel,
    QComboBox,
    QMenu,
    QSpacerItem,
    QSizePolicy,
)

from rare.models.options import options, LibraryFilter, LibraryOrder
from rare.shared import RareCore
from rare.utils.misc import qta_icon
from rare.widgets.button_edit import ButtonLineEdit


class LibraryHeadBar(QWidget):
    filterChanged = Signal(object)
    orderChanged = Signal(object)
    viewChanged = Signal(object)
    goto_import = Signal()
    goto_egl_sync = Signal()
    goto_eos_ubisoft = Signal()

    def __init__(self, parent=None):
        super(LibraryHeadBar, self).__init__(parent=parent)
        self.logger = logging.getLogger(type(self).__name__)
        self.rcore = RareCore.instance()
        self.settings = QSettings(self)

        self.filter = QComboBox(self)
        filters = {
            LibraryFilter.ALL: self.tr("All games"),
            LibraryFilter.INSTALLED: self.tr("Installed"),
            LibraryFilter.OFFLINE: self.tr("Offline"),
            # LibraryFilter.HIDDEN: self.tr("Hidden"),
        }
        for data, text in filters.items():
            self.filter.addItem(text, data)

        if self.rcore.bit32_games:
            self.filter.addItem(self.tr("32bit games"), LibraryFilter.WIN32)
        if self.rcore.mac_games:
            self.filter.addItem(self.tr("macOS games"), LibraryFilter.MAC)
        if self.rcore.origin_games:
            self.filter.addItem(self.tr("Exclude Origin"), LibraryFilter.INSTALLABLE)
        self.filter.addItem(self.tr("Include Unreal"),     LibraryFilter.INCLUDE_UE)

        try:
            _filter = LibraryFilter(self.settings.value(*options.library_filter))
            if (index := self.filter.findData(_filter, Qt.ItemDataRole.UserRole)) < 0:
                raise ValueError(f"Filter '{_filter}' is not available")
            else:
                self.filter.setCurrentIndex(index)
        except (TypeError, ValueError) as e:
            self.logger.error("Error while loading library: %s", e)
            self.settings.setValue(options.library_filter.key, options.library_filter.default)
            _filter = LibraryFilter(options.library_filter.default)
            self.filter.setCurrentIndex(self.filter.findData(_filter, Qt.ItemDataRole.UserRole))
        self.filter.currentIndexChanged.connect(self.__filter_changed)

        self.order = QComboBox(parent=self)
        sortings = {
            LibraryOrder.TITLE:  self.tr("Title"),
            LibraryOrder.RECENT: self.tr("Recently played"),
            LibraryOrder.NEWEST: self.tr("Newest"),
            LibraryOrder.OLDEST: self.tr("Oldest"),
        }
        for data, text in sortings.items():
            self.order.addItem(text, data)

        try:
            _order = LibraryOrder(self.settings.value(*options.library_order))
            if (index := self.order.findData(_order, Qt.ItemDataRole.UserRole)) < 0:
                raise ValueError(f"Order '{_order}' is not available")
            else:
                self.order.setCurrentIndex(index)
        except (TypeError, ValueError) as e:
            self.logger.error("Error while loading library: %s", e)
            self.settings.setValue(options.library_order.key, options.library_order.default)
            _order = LibraryOrder(options.library_order.default)
            self.order.setCurrentIndex(self.order.findData(_order, Qt.ItemDataRole.UserRole))
        self.order.currentIndexChanged.connect(self.__order_changed)

        integrations_menu = QMenu(parent=self)
        import_action = QAction(
            qta_icon("mdi.import", "fa.arrow-down"), self.tr("Import Game"), integrations_menu
        )

        import_action.triggered.connect(self.goto_import)
        egl_sync_action = QAction(qta_icon("mdi.sync", "fa.refresh"), self.tr("Sync with EGL"), integrations_menu)
        egl_sync_action.triggered.connect(self.goto_egl_sync)

        eos_ubisoft_action = QAction(
            qta_icon("mdi.rocket", "fa.rocket"), self.tr("Epic Overlay and Ubisoft"), integrations_menu
        )
        eos_ubisoft_action.triggered.connect(self.goto_eos_ubisoft)

        integrations_menu.addAction(import_action)
        integrations_menu.addAction(egl_sync_action)
        integrations_menu.addAction(eos_ubisoft_action)

        integrations = QPushButton(parent=self)
        integrations.setText(self.tr("Integrations"))
        integrations.setIcon(qta_icon("mdi.tools"))
        integrations.setMenu(integrations_menu)

        self.search_bar = ButtonLineEdit("fa.search", placeholder_text=self.tr("Search"))
        self.search_bar.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Preferred)
        self.search_bar.setObjectName("SearchBar")
        self.search_bar.setMinimumWidth(250)

        installed_tooltip = self.tr("Installed games")
        self.installed_icon = QLabel(parent=self)
        self.installed_icon.setPixmap(qta_icon("ph.floppy-disk-back-fill").pixmap(QSize(16, 16)))
        self.installed_icon.setToolTip(installed_tooltip)
        self.installed_label = QLabel(parent=self)
        font = self.installed_label.font()
        font.setBold(True)
        self.installed_label.setFont(font)
        self.installed_label.setToolTip(installed_tooltip)
        available_tooltip = self.tr("Available games")
        self.available_icon = QLabel(parent=self)
        self.available_icon.setPixmap(qta_icon("ph.floppy-disk-back-light").pixmap(QSize(16, 16)))
        self.available_icon.setToolTip(available_tooltip)
        self.available_label = QLabel(parent=self)
        self.available_label.setToolTip(available_tooltip)

        self.refresh_list = QPushButton(parent=self)
        self.refresh_list.setIcon(qta_icon("fa.refresh"))  # Reload icon
        self.refresh_list.clicked.connect(self.__refresh_clicked)

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.filter)
        layout.addWidget(self.order)
        layout.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed))
        layout.addWidget(self.search_bar)
        layout.addWidget(self.installed_icon)
        layout.addWidget(self.installed_label)
        layout.addWidget(self.available_icon)
        layout.addWidget(self.available_label)
        layout.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed))
        layout.addWidget(integrations)
        layout.addWidget(self.refresh_list)

    def set_games_count(self, inst: int, avail: int) -> None:
        self.installed_label.setText(str(inst))
        self.available_label.setText(str(avail))

    @Slot()
    def __refresh_clicked(self):
        self.rcore.fetch()

    def current_filter(self) -> LibraryFilter:
        return self.filter.currentData(Qt.ItemDataRole.UserRole)

    @Slot(int)
    def __filter_changed(self, index: int):
        data = self.filter.itemData(index, Qt.ItemDataRole.UserRole)
        self.filterChanged.emit(data)
        self.settings.setValue(options.library_filter.key, int(data))

    def current_order(self) -> LibraryOrder:
        return self.order.currentData(Qt.ItemDataRole.UserRole)

    @Slot(int)
    def __order_changed(self, index: int):
        data = self.order.itemData(index, Qt.ItemDataRole.UserRole)
        self.orderChanged.emit(data)
        self.settings.setValue(options.library_order.key, int(data))


class SelectViewWidget(QWidget):
    toggled = Signal(bool)

    def __init__(self, icon_view: bool, parent=None):
        super(SelectViewWidget, self).__init__(parent=parent)
        self.icon_button = QPushButton(self)
        self.icon_button.setObjectName(f"{type(self).__name__}Button")
        self.list_button = QPushButton(self)
        self.list_button.setObjectName(f"{type(self).__name__}Button")

        if icon_view:
            self.icon_button.setIcon(qta_icon("mdi.view-grid-outline", "ei.th-large", color="orange"))
            self.list_button.setIcon(qta_icon("fa5s.list", "ei.th-list", color="#eee"))
        else:
            self.icon_button.setIcon(qta_icon("mdi.view-grid-outline", "ei.th-large", color="#eee"))
            self.list_button.setIcon(qta_icon("fa5s.list", "ei.th-list", color="orange"))

        self.icon_button.clicked.connect(self.icon)
        self.list_button.clicked.connect(self.list)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.icon_button)
        layout.addWidget(self.list_button)

        self.setLayout(layout)

    def icon(self):
        self.icon_button.setIcon(qta_icon("mdi.view-grid-outline", "ei.th-large", color="orange"))
        self.list_button.setIcon(qta_icon("fa5s.list", "ei.th-list", color="#eee"))
        self.toggled.emit(True)

    def list(self):
        self.icon_button.setIcon(qta_icon("mdi.view-grid-outline", "ei.th-large", color="#eee"))
        self.list_button.setIcon(qta_icon("fa5s.list", "ei.th-list", color="orange"))
        self.toggled.emit(False)






FEED_QUERY = '''
query feedQuery(
  $locale: String!
  $countryCode: String
  $offset: Int
  $postsPerPage: Int
  $category: String
) {
  TransientStream {
    myTransientFeed(countryCode: $countryCode, locale: $locale) {
      id
      activity {
        ... on LinkAccountActivity {
          type
          created_at
          platforms
        }
        ... on SuggestedFriendsActivity {
          type
          created_at
          platform
          suggestions {
            epicId
            epicDisplayName
            platformFullName
            platformAvatar
          }
        }
        ... on IncomingInvitesActivity {
          type
          created_at
          invites {
            epicId
            epicDisplayName
          }
        }
        ... on RecentPlayersActivity {
          type
          created_at
          players {
            epicId
            epicDisplayName
            playedGameName
          }
        }
      }
    }
  }
  Blog {
    dieselBlogPosts: getPosts(
      locale: $locale
      offset: $offset
      postsPerPage: $postsPerPage
      category: $category
    ) {
      blogList {
        _id
        author
        category
        content
        urlPattern
        slug
        sticky
        title
        date
        image
        shareImage
        trendingImage
        url
        featured
        link
        externalLink
      }
    }
  }
}
'''

REVIEWS_QUERY = '''
query productReviewsQuery($sku: String!) {
  OpenCritic {
    productReviews(sku: $sku) {
      id
      name
      openCriticScore
      reviewCount
      percentRecommended
      openCriticUrl
      award
      topReviews {
        publishedDate
        externalUrl
        snippet
        language
        score
        author
        ScoreFormat {
          id
          description
        }
        OutletId
        outletName
        displayScore
      }
    }
  }
}
'''

MEDIA_QUERY = '''
query fetchMediaRef($mediaRefId: String!) {
  Media {
    getMediaRef(mediaRefId: $mediaRefId) {
      accountId
      outputs {
        duration
        url
        width
        height
        key
        contentType
      }
      namespace
    }
  }
}
'''

ADDONS_QUERY = '''
query getAddonsByNamespace(
  $categories: String!
  $count: Int!
  $country: String!
  $locale: String!
  $namespace: String!
  $sortBy: String!
  $sortDir: String!
) {
  Catalog {
    catalogOffers(
      namespace: $namespace
      locale: $locale
      params: {
        category: $categories
        count: $count
        country: $country
        sortBy: $sortBy
        sortDir: $sortDir
      }
    ) {
      elements {
        countriesBlacklist
        customAttributes {
          key
          value
        }
        description
        developer
        effectiveDate
        id
        isFeatured
        keyImages {
          type
          url
        }
        lastModifiedDate
        longDescription
        namespace
        offerType
        productSlug
        releaseDate
        status
        technicalDetails
        title
        urlSlug
      }
    }
  }
}
'''

CATALOG_QUERY = '''
query catalogQuery(
  $category: String
  $count: Int
  $country: String!
  $keywords: String
  $locale: String
  $namespace: String!
  $sortBy: String
  $sortDir: String
  $start: Int
  $tag: String
) {
  Catalog {
    catalogOffers(
      namespace: $namespace
      locale: $locale
      params: {
        count: $count
        country: $country
        category: $category
        keywords: $keywords
        sortBy: $sortBy
        sortDir: $sortDir
        start: $start
        tag: $tag
      }
    ) {
      elements {
        isFeatured
        collectionOfferIds
        title
        id
        namespace
        description
        keyImages {
          type
          url
        }
        seller {
          id
          name
        }
        productSlug
        urlSlug
        items {
          id
          namespace
        }
        customAttributes {
          key
          value
        }
        categories {
          path
        }
        price(country: $country) {
          totalPrice {
            discountPrice
            originalPrice
            voucherDiscount
            discount
            fmtPrice(locale: $locale) {
              originalPrice
              discountPrice
              intermediatePrice
            }
          }
          lineOffers {
            appliedRules {
              id
              endDate
            }
          }
        }
        linkedOfferId
        linkedOffer {
          effectiveDate
          customAttributes {
            key
            value
          }
        }
      }
      paging {
        count
        total
      }
    }
  }
}
'''

CATALOG_TAGS_QUERY = '''
query catalogTags($namespace: String!) {
  Catalog {
    tags(namespace: $namespace, start: 0, count: 999) {
      elements {
        aliases
        id
        name
        referenceCount
        status
      }
    }
  }
}
'''

PREREQUISITES_QUERY = '''
query fetchPrerequisites($offerParams: [OfferParams]) {
  Launcher {
    prerequisites(offerParams: $offerParams) {
      namespace
      offerId
      missingPrerequisiteItems
      satisfiesPrerequisites
    }
  }
}
'''

PROMOTIONS_QUERY = '''
query promotionsQuery(
  $namespace: String!
  $country: String!
  $locale: String!
) {
  Catalog {
    catalogOffers(
      namespace: $namespace
      locale: $locale
      params: {
        category: "freegames"
        country: $country
        sortBy: "effectiveDate"
        sortDir: "asc"
      }
    ) {
      elements {
        title
        description
        id
        namespace
        categories {
          path
        }
        linkedOfferNs
        linkedOfferId
        keyImages {
          type
          url
        }
        productSlug
        promotions {
          promotionalOffers {
            promotionalOffers {
              startDate
              endDate
              discountSetting {
                discountType
                discountPercentage
              }
            }
          }
          upcomingPromotionalOffers {
            promotionalOffers {
              startDate
              endDate
              discountSetting {
                discountType
                discountPercentage
              }
            }
          }
        }
      }
    }
  }
}
'''

OFFERS_QUERY = '''
query catalogQuery(
  $productNamespace: String!
  $offerId: String!
  $locale: String
  $country: String!
  $includeSubItems: Boolean!
) {
  Catalog {
    catalogOffer(namespace: $productNamespace, id: $offerId, locale: $locale) {
      title
      id
      namespace
      description
      effectiveDate
      expiryDate
      isCodeRedemptionOnly
      keyImages {
        type
        url
      }
      seller {
        id
        name
      }
      productSlug
      urlSlug
      url
      tags {
        id
      }
      items {
        id
        namespace
      }
      customAttributes {
        key
        value
      }
      categories {
        path
      }
      price(country: $country) {
        totalPrice {
          discountPrice
          originalPrice
          voucherDiscount
          discount
          currencyCode
          currencyInfo {
            decimals
          }
          fmtPrice(locale: $locale) {
            originalPrice
            discountPrice
            intermediatePrice
          }
        }
        lineOffers {
          appliedRules {
            id
            endDate
            discountSetting {
              discountType
            }
          }
        }
      }
    }
    offerSubItems(namespace: $productNamespace, id: $offerId)
      @include(if: $includeSubItems) {
      namespace
      id
      releaseInfo {
        appId
        platform
      }
    }
  }
}
'''

SEARCH_STORE_QUERY = '''
query searchStoreQuery(
  $allowCountries: String
  $category: String
  $count: Int
  $country: String!
  $keywords: String
  $locale: String
  $namespace: String
  $itemNs: String
  $sortBy: String
  $sortDir: String
  $start: Int
  $tag: String
  $releaseDate: String
  $withPrice: Boolean = false
  $withPromotions: Boolean = false
) {
  Catalog {
    searchStore(
      allowCountries: $allowCountries
      category: $category
      count: $count
      country: $country
      keywords: $keywords
      locale: $locale
      namespace: $namespace
      itemNs: $itemNs
      sortBy: $sortBy
      sortDir: $sortDir
      releaseDate: $releaseDate
      start: $start
      tag: $tag
    ) {
      elements {
        title
        id
        namespace
        description
        effectiveDate
        keyImages {
          type
          url
        }
        seller {
          id
          name
        }
        productSlug
        urlSlug
        url
        tags {
          id
        }
        items {
          id
          namespace
        }
        customAttributes {
          key
          value
        }
        categories {
          path
        }
        price(country: $country) @include(if: $withPrice) {
          totalPrice {
            discountPrice
            originalPrice
            voucherDiscount
            discount
            currencyCode
            currencyInfo {
              decimals
            }
            fmtPrice(locale: $locale) {
              originalPrice
              discountPrice
              intermediatePrice
            }
          }
          lineOffers {
            appliedRules {
              id
              endDate
              discountSetting {
                discountType
              }
            }
          }
        }
        promotions(category: $category) @include(if: $withPromotions) {
          promotionalOffers {
            promotionalOffers {
              startDate
              endDate
              discountSetting {
                discountType
                discountPercentage
              }
            }
          }
          upcomingPromotionalOffers {
            promotionalOffers {
              startDate
              endDate
              discountSetting {
                discountType
                discountPercentage
              }
            }
          }
        }
      }
      paging {
        count
        total
      }
    }
  }
}
'''


import os
from logging import getLogger
from typing import Tuple, Union, Optional

from PySide6.QtCore import Signal, Qt
from PySide6.QtGui import QShowEvent
from PySide6.QtWidgets import QGroupBox, QFileDialog, QFormLayout, QComboBox

from rare.models.wrapper import Wrapper, WrapperType
from rare.shared import RareCore
from rare.shared.wrappers import Wrappers
from rare.utils import config_helper as config
from rare.utils.compat import steam
from rare.utils.paths import proton_compat_dir
from rare.widgets.indicator_edit import PathEdit, IndicatorReasonsCommon

logger = getLogger("ProtonSettings")


class ProtonSettings(QGroupBox):
    # str: option key
    environ_changed: Signal = Signal(str)
    # bool: state
    tool_enabled: Signal = Signal(bool)

    def __init__(self, parent=None):
        super(ProtonSettings, self).__init__(parent=parent)
        self.setTitle(self.tr("Proton settings"))

        self.tool_combo = QComboBox(self)
        self.tool_combo.currentIndexChanged.connect(self.__on_proton_changed)

        self.tool_prefix = PathEdit(
            file_mode=QFileDialog.FileMode.Directory,
            edit_func=self.proton_prefix_edit,
            save_func=self.proton_prefix_save,
            placeholder=self.tr("Please select path for proton prefix"),
            parent=self
        )

        layout = QFormLayout(self)
        layout.addRow(self.tr("Proton tool"), self.tool_combo)
        layout.addRow(self.tr("Compat data"), self.tool_prefix)
        layout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        layout.setLabelAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        layout.setFormAlignment(Qt.AlignmentFlag.AlignLeading | Qt.AlignmentFlag.AlignTop)

        self.app_name: str = "default"
        self.core = RareCore.instance().core()
        self.wrappers: Wrappers = RareCore.instance().wrappers()
        self.tool_wrapper: Optional[Wrapper] = None

    def showEvent(self, a0: QShowEvent) -> None:
        if a0.spontaneous():
            return super().showEvent(a0)

        self.tool_combo.blockSignals(True)
        self.tool_combo.clear()
        self.tool_combo.addItem(self.tr("Don't use a compatibility tool"), None)
        tools = steam.find_tools()
        for tool in tools:
            self.tool_combo.addItem(tool.name, tool)
        try:
            wrapper = next(
                filter(lambda w: w.is_compat_tool, self.wrappers.get_game_wrapper_list(self.app_name))
            )
            self.tool_wrapper = wrapper
            tool = next(filter(lambda t: t.checksum == wrapper.checksum, tools))
            index = self.tool_combo.findData(tool)
        except StopIteration:
            index = 0
        self.tool_combo.setCurrentIndex(index)
        self.tool_combo.blockSignals(False)

        enabled = bool(self.tool_combo.currentData(Qt.ItemDataRole.UserRole))
        self.tool_prefix.blockSignals(True)
        self.tool_prefix.setText(config.get_proton_compatdata(self.app_name, fallback=""))
        self.tool_prefix.setEnabled(enabled)
        self.tool_prefix.blockSignals(False)

        super().showEvent(a0)

    def __on_proton_changed(self, index):
        steam_tool: Union[steam.ProtonTool, steam.CompatibilityTool] = self.tool_combo.itemData(index)

        steam_environ = steam.get_steam_environment(steam_tool, self.tool_prefix.text())
        for key, value in steam_environ.items():
            config.save_envvar(self.app_name, key, value)
            self.environ_changed.emit(key)

        wrappers = self.wrappers.get_game_wrapper_list(self.app_name)
        if self.tool_wrapper and self.tool_wrapper in wrappers:
            wrappers.remove(self.tool_wrapper)
        if steam_tool is None:
            self.tool_wrapper = None
        else:
            wrapper = Wrapper(
                command=steam_tool.command(), name=steam_tool.name, wtype=WrapperType.COMPAT_TOOL
            )
            wrappers.append(wrapper)
            self.tool_wrapper = wrapper
        self.wrappers.set_game_wrapper_list(self.app_name, wrappers)

        self.tool_prefix.setEnabled(steam_tool is not None)
        if steam_tool:
            if not (compatdata_path := config.get_proton_compatdata(self.app_name, fallback="")):
                compatdata_path = proton_compat_dir(self.app_name)
                config.save_proton_compatdata(self.app_name, str(compatdata_path))
                target = compatdata_path.joinpath("pfx")
                if not target.is_dir():
                    os.makedirs(target, exist_ok=True)
            self.tool_prefix.setText(str(compatdata_path))
        else:
            self.tool_prefix.setText("")

        self.tool_enabled.emit(steam_tool is not None)

    @staticmethod
    def proton_prefix_edit(text: str) -> Tuple[bool, str, int]:
        if not text:
            return False, text, IndicatorReasonsCommon.EMPTY
        parent_dir = os.path.dirname(text)
        return os.path.exists(parent_dir), text, IndicatorReasonsCommon.DIR_NOT_EXISTS

    def proton_prefix_save(self, text: str):
        if not text:
            return
        config.save_proton_compatdata(self.app_name, text)
        self.environ_changed.emit("STEAM_COMPAT_DATA_PATH")




from cx_Freeze import setup, Executable

from rare import __version__ as _version

_name = "Rare"
_author = "RareDevs"
_description = "Open source alternative for Epic Games Launcher, using Legendary"

_icon = "rare/resources/images/Rare"
_license = "LICENSE"

build_exe_options = {
    "bin_excludes": [ "qpdf.dll", "libqpdf.so", "libqpdf.dylib"],
    "excludes": [
        "tkinter",
        "unittest",
        "pydoc",
    ],
    "include_msvcr": True,
    "optimize": 2,
}

shortcut_table = [
    (
        "DesktopShortcut",  # Shortcut
        "DesktopFolder",  # Directory_
        "Rare",  # Name
        "TARGETDIR",  # Component_
        "[TARGETDIR]Rare.exe",  # Target
        None,  # Arguments
        None,  # Description
        None,  # Hotkey
        None,  # Icon
        None,  # IconIndex
        None,  # ShowCmd
        "TARGETDIR",  # WkDir
    ),
]

msi_data = {
    "Shortcut": shortcut_table,
    "ProgId": [
        ("Prog.Id", None, None, _description, "IconId", None),
    ],
    "Icon": [("IcodId", f"{_icon}.ico")],
}

bdist_msi_options = {
    "data": msi_data,
    "license_file": _license,
    # generated with str(uuid.uuid3(uuid.NAMESPACE_DNS, 'io.github.dummerle.rare')).upper()
    "upgrade_code": "{85D9FCC2-733E-3D74-8DD4-8FE33A07ADF8}",
}

bdist_appimage_options = {
    "target_name": _name,
    "target_version": _version,
}

bdist_mac_options = {
    "iconfile": f"{_icon}.icns",
    "bundle_name": f"{_name}",
}

bdist_dmg_options = {
    "volume_label": _name,
    "applications_shortcut": True,
    "show_icon_preview": True,
    "license": {
        "default-language": "en_US",
        "licenses": {
            "en_US": _license,
        },
        "buttons": {
            "en_US": [
                "English", "Agree", "Disagree", "Print", "Save", "License"
            ]
        },
    },
}

executables = [
    Executable(
        script="rare/main.py",
        copyright=f"Copyright (C) 2024 {_author}",
        base="gui",
        icon=_icon,
        target_name=_name,
    ),
]

setup(
    name=_name,
    version=_version,
    author=_author,
    description=_description,
    executables=executables,
    options={
        "build_exe": build_exe_options,
        "bdist_msi": bdist_msi_options,
        "bdist_appimage": bdist_appimage_options,
        "bdist_mac": bdist_mac_options,
        "bdist_dmg": bdist_dmg_options,
    },
)



import hashlib
import json
import pickle
import zlib

# from concurrent import futures
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Set
from typing import Tuple, Dict, Union, Type, List, Callable

import requests
from PySide6.QtCore import Qt, Signal, QObject, QSize, QThreadPool, QRunnable, QRect, QRectF, Slot
from PySide6.QtGui import QPixmap, QImage, QPainter, QPainterPath, QBrush, QTransform, QPen
from PySide6.QtWidgets import QApplication
from legendary.models.game import Game

from rare.lgndr.core import LegendaryCore
from rare.models.image import ImageSize, ImageType
from rare.models.signals import GlobalSignals
from rare.utils.paths import (
    image_dir,
    image_dir_game,
    image_tall_path,
    image_wide_path,
    image_icon_path,
    resources_path,
    desktop_icon_suffix,
    desktop_icon_path,
)

# from requests_futures.sessions import FuturesSession

if TYPE_CHECKING:
    pass

logger = getLogger("ImageManager")


class ImageManager(QObject):
    class Worker(QRunnable):
        class Signals(QObject):
            # object: Game
            completed = Signal(object)

        def __init__(self, func: Callable, updates: List, json_data: Dict, game: Game):
            super(ImageManager.Worker, self).__init__()
            self.signals = ImageManager.Worker.Signals()
            self.setAutoDelete(True)
            self.func = func
            self.updates = updates
            self.json_data = json_data
            self.game = game

        def run(self):
            self.func(self.updates, self.json_data, self.game)
            logger.debug(f"Emitting singal for {self.game.app_name} ({self.game.app_title})")
            self.signals.completed.emit(self.game)

    def __init__(self, signals: GlobalSignals, core: LegendaryCore):
        # lk: the ordering in __img_types matters for the order of fallbacks
        # {'AndroidIcon', 'DieselGameBox', 'DieselGameBoxLogo', 'DieselGameBoxTall', 'DieselGameBoxWide',
        #  'ESRB', 'Featured', 'OfferImageTall', 'OfferImageWide', 'Screenshot', 'Thumbnail'}
        self.__img_tall_types: Tuple = ("DieselGameBoxTall", "OfferImageTall", "Thumbnail")
        self.__img_wide_types: Tuple = ("DieselGameBoxWide", "DieselGameBox", "OfferImageWide", "Screenshot")
        self.__img_logo_types: Tuple = ("DieselGameBoxLogo",)
        self.__img_types: Tuple = self.__img_tall_types + self.__img_wide_types + self.__img_logo_types
        self.__dl_retries = 1
        self.__worker_app_names: Set[str] = set()
        super(ImageManager, self).__init__()
        self.signals = signals
        self.core = core

        self.image_dir: Path = image_dir()
        if not self.image_dir.is_dir():
            self.image_dir.mkdir()
            logger.info(f"Created image directory at {self.image_dir}")

        self.threadpool = QThreadPool()
        self.threadpool.setMaxThreadCount(6)

    @staticmethod
    def __img_json(app_name: str) -> Path:
        return image_dir_game(app_name).joinpath("image.json")

    @staticmethod
    def __img_cache(app_name: str) -> Path:
        return image_dir_game(app_name).joinpath("image.cache")

    @staticmethod
    def __img_all(app_name: str) -> Tuple:
        return (
            image_tall_path(app_name),
            image_tall_path(app_name, color=False),
            image_wide_path(app_name),
            image_wide_path(app_name, color=False),
            image_icon_path(app_name),
            image_icon_path(app_name, color=False),
            desktop_icon_path(app_name),
        )

    def __prepare_download(self, game: Game, force: bool = False) -> Tuple[List, Dict]:
        if force and image_dir_game(game.app_name).exists():
            for file in self.__img_all(game.app_name):
                file.unlink(missing_ok=True)
        if not image_dir_game(game.app_name).is_dir():
            image_dir_game(game.app_name).mkdir()

        # Load image checksums
        if not self.__img_json(game.app_name).is_file():
            json_data: Dict = dict(zip(self.__img_types, [None] * len(self.__img_types)))
        else:
            json_data = json.load(open(self.__img_json(game.app_name), "r"))

        # Only download the best matching candidate for each image category
        def best_match(key_images: List, image_types: Tuple) -> Dict:
            matches = sorted(
                filter(lambda image: image["type"] in image_types, key_images),
                key=lambda x: image_types.index(x["type"]) if x["type"] in image_types else len(image_types),
                reverse=False,
            )
            try:
                best = matches[0]
            except IndexError as e:
                best = {}
            return best

        candidates = tuple(
            image
            for image in [
                best_match(game.metadata.get("keyImages", []), self.__img_tall_types),
                best_match(game.metadata.get("keyImages", []), self.__img_wide_types),
                best_match(game.metadata.get("keyImages", []), self.__img_logo_types),
            ]
            if bool(image)
        )

        # lk: Find updates or initialize if images are missing.
        # lk: `updates` will be empty for games without images
        # lk: so everything below it is skipped
        # TODO: Move this into the thread, maybe, concurrency could help here too
        updates = []
        if not all(file.is_file() for file in self.__img_all(game.app_name)):
            # lk: fast path for games without images, convert Rare's logo
            if not candidates:
                cache_data: Dict = dict(zip(self.__img_types, [None] * len(self.__img_types)))
                with open(resources_path.joinpath("images", "cover.png"), "rb") as fd:
                    cache_data["DieselGameBoxTall"] = fd.read()
                    fd.seek(0)
                    cache_data["DieselGameBoxWide"] = fd.read()
                # cache_data["DieselGameBoxLogo"] = open(
                #         resources_path.joinpath("images", "logo.png"), "rb").read()
                self.__convert(game, cache_data)
                json_data["cache"] = None
                json_data["scale"] = ImageSize.Tall.pixel_ratio
                json_data["size"] = {"w": ImageSize.Tall.size.width(), "h": ImageSize.Tall.size.height()}
                with open(self.__img_json(game.app_name), "w", encoding="utf-8") as file:
                    json.dump(json_data, file)
            else:
                updates = [image for image in candidates if image["type"] in self.__img_types]
        else:
            for image in candidates:
                if image["type"] in self.__img_types:
                    if image["type"] not in json_data.keys() or json_data[image["type"]] != image["md5"]:
                        updates.append(image)

        return updates, json_data

    def __download(self, updates: List, json_data: Dict, game: Game, use_async: bool = False) -> bool:
        # Decompress existing image.cache
        if not self.__img_cache(game.app_name).is_file():
            cache_data = dict(zip(self.__img_types, [None] * len(self.__img_types)))
        else:
            cache_data = self.__decompress(game)

        # lk: filter updates again against the cache now that it is available
        # images in cache don't need to be downloaded again.
        downloads = [
            image
            for image in updates
            if (cache_data.get(image["type"], None) is None or json_data[image["type"]] != image["md5"])
        ]

        # Download
        # # lk: Keep this here, so I don't have to go looking for it again,
        # # lk: it might be useful in the future.
        # if use_async:
        #     session = FuturesSession(max_workers=len(self.__img_types))
        #     image_requests = []
        #     for image in downloads:
        #         logger.info(f"Downloading {image['type']} for {game.app_title}")
        #         json_data[image["type"]] = image["md5"]
        #         if image["type"] in self.__img_tall_types:
        #             payload = {"resize": 1, "w": ImageSize.Image.size.width(), "h": ImageSize.Image.size.height()}
        #         elif image["type"] in self.__img_wide_types:
        #             payload = {"resize": 1, "w": ImageSize.ImageWide.size.width(), "h": ImageSize.ImageWide.size.height()}
        #         else:
        #             # Set the larger of the sizes for everything else
        #             payload = {"resize": 1, "w": ImageSize.ImageWide.size.width(), "h": ImageSize.ImageWide.size.height()}
        #         req = session.get(image["url"], params=payload)
        #         req.image_type = image["type"]
        #         image_requests.append(req)
        #     for req in futures.as_completed(image_requests):
        #         cache_data[req.image_type] = req.result().content
        # else:
        for image in downloads:
            logger.info(f"Downloading {image['type']} for {game.app_name} ({game.app_title})")
            json_data[image["type"]] = image["md5"]
            if image["type"] in self.__img_tall_types:
                payload = {"resize": 1, "w": ImageSize.Tall.size.width(), "h": ImageSize.Tall.size.height()}
            elif image["type"] in self.__img_wide_types:
                payload = {"resize": 1, "w": ImageSize.Wide.size.width(), "h": ImageSize.Wide.size.height()}
            else:
                # Set the larger of the sizes for everything else
                payload = {"resize": 1, "w": ImageSize.Wide.size.width(), "h": ImageSize.Wide.size.height()}
            try:
                # cache_data[image["type"]] = requests.get(image["url"], params=payload).content
                cache_data[image["type"]] = requests.get(image["url"], params=payload, timeout=10).content
            except Exception as e:
                logger.error(e)
                return False

        self.__convert(game, cache_data)
        # lk: don't keep the cache if there is no logo (kept for me)
        # if cache_data["DieselGameBoxLogo"] is not None:
        #     self.__compress(game, cache_data)
        self.__compress(game, cache_data)

        # hash image cache
        try:
            with open(self.__img_cache(game.app_name), "rb") as archive:
                archive_hash = hashlib.md5(archive.read()).hexdigest()
        except FileNotFoundError:
            archive_hash = None

        json_data["cache"] = archive_hash
        json_data["scale"] = ImageSize.Tall.pixel_ratio
        json_data["size"] = {"w": ImageSize.Tall.size.width(), "h": ImageSize.Tall.size.height()}

        # write image.json
        with open(self.__img_json(game.app_name), "w", encoding="utf-8") as file:
            json.dump(json_data, file)

        return bool(updates)

    __icon_overlay: Optional[QPainterPath] = None

    @staticmethod
    def __generate_icon_overlay(rect: QRect) -> QPainterPath:
        if ImageManager.__icon_overlay is not None:
            return ImageManager.__icon_overlay
        rounded_path = QPainterPath()
        margin = 0.05
        rounded_path.addRoundedRect(
            QRectF(
                rect.width() * margin,
                rect.height() * margin,
                rect.width() - (rect.width() * margin * 2),
                rect.height() - (rect.width() * margin * 2),
            ),
            rect.height() * 0.2,
            rect.height() * 0.2,
        )
        ImageManager.__icon_overlay = rounded_path
        return ImageManager.__icon_overlay

    @staticmethod
    def __convert_icon(cover: QImage) -> QImage:
        icon_size = QSize(128, 128)
        icon = QImage(icon_size, QImage.Format.Format_ARGB32_Premultiplied)
        painter = QPainter(icon)
        painter.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform, True)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
        painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_Source)
        painter.fillRect(icon.rect(), Qt.GlobalColor.transparent)
        overlay = ImageManager.__generate_icon_overlay(icon.rect())
        brush = QBrush(cover)
        scale = max(icon.width() / cover.width(), icon.height() / cover.height())
        transform = QTransform().scale(scale, scale)
        brush.setTransform(transform)
        painter.fillPath(overlay, brush)
        pen = QPen(Qt.GlobalColor.black, 2)
        painter.setPen(pen)
        painter.drawPath(overlay)
        painter.end()
        return icon

    def __convert(self, game, images, force=False) -> None:
        for file in self.__img_all(game.app_name):
            if force and file.exists():
                file.unlink(missing_ok=True)

        def find_image_data(image_types: Tuple):
            data = None
            for image_type in image_types:
                if images.get(image_type, None) is not None:
                    data = images[image_type]
                    break
            return data

        tall_data = find_image_data(self.__img_tall_types)
        wide_data = find_image_data(self.__img_wide_types)
        logo_data = find_image_data(self.__img_logo_types)

        def convert_image(image_data, logo_data, preset: ImageSize.Preset) -> QImage:
            image = QImage()
            image.loadFromData(image_data)
            image.convertToFormat(QImage.Format.Format_ARGB32_Premultiplied)
            # lk: Images are not always at the correct aspect ratio, so crop them to size
            wr, hr = preset.aspect_ratio
            factor = min(image.width() // wr, image.height() // hr)
            rem_w = (image.width() - factor * wr) // 2
            rem_h = (image.height() - factor * hr) // 2
            image = image.copy(rem_w, rem_h, factor * wr, factor * hr)

            if logo_data is not None:
                logo = QImage()
                logo.loadFromData(logo_data)
                logo.convertToFormat(QImage.Format.Format_ARGB32_Premultiplied)
                if logo.width() > image.width():
                    logo = logo.scaled(
                        image.width(),
                        image.height(),
                        Qt.AspectRatioMode.KeepAspectRatio,
                        Qt.TransformationMode.SmoothTransformation
                    )
                painter = QPainter(image)
                painter.drawImage((image.width() - logo.width()) // 2, image.height() - logo.height(), logo)
                painter.end()

            return image.scaled(
                preset.size,
                Qt.AspectRatioMode.KeepAspectRatio,
                Qt.TransformationMode.SmoothTransformation
            )

        tall = convert_image(tall_data, logo_data, ImageSize.Tall)
        wide = convert_image(wide_data, logo_data, ImageSize.Wide)

        icon = self.__convert_icon(tall)
        icon.save(desktop_icon_path(game.app_name).as_posix(), format=desktop_icon_suffix().upper())

        def save_image(image: QImage, color_path: Path, gray_path: Path):
            # this is not required if we ever want to re-apply the alpha channel
            # image = image.convertToFormat(QImage.Format_Indexed8)
            # add the alpha channel back to the cover
            image = image.convertToFormat(QImage.Format.Format_ARGB32_Premultiplied)
            image.save(color_path.as_posix(), format="PNG")
            # quick way to convert to grayscale, but keep the alpha channel
            alpha = image.convertToFormat(QImage.Format.Format_Alpha8)
            image = image.convertToFormat(QImage.Format.Format_Grayscale8)
            # add the alpha channel back to the grayscale cover
            image = image.convertToFormat(QImage.Format.Format_ARGB32_Premultiplied)
            image.setAlphaChannel(alpha)
            image.save(gray_path.as_posix(), format="PNG")

        save_image(icon, image_icon_path(game.app_name), image_icon_path(game.app_name, color=False))
        save_image(tall, image_tall_path(game.app_name), image_tall_path(game.app_name, color=False))
        save_image(wide, image_wide_path(game.app_name), image_wide_path(game.app_name, color=False))

    def __compress(self, game: Game, data: Dict) -> None:
        archive = open(self.__img_cache(game.app_name), "wb")
        cdata = zlib.compress(pickle.dumps(data), level=-1)
        archive.write(cdata)
        archive.close()

    def __decompress(self, game: Game) -> Dict:
        archive = open(self.__img_cache(game.app_name), "rb")
        try:
            data = zlib.decompress(archive.read())
            data = pickle.loads(data)
        except zlib.error:
            data = dict(zip(self.__img_types, [None] * len(self.__img_types)))
        finally:
            archive.close()
        return data

    def __append_to_queue(self, game: Game):
        self.__worker_app_names.add(game.app_name)

    @Slot(object)
    def __remove_from_queue(self, game: Game):
        self.__worker_app_names.remove(game.app_name)

    def download_image(
            self, game: Game, load_callback: Callable[[], None], priority: int, force: bool = False
    ) -> None:
        if game.app_name in self.__worker_app_names:
            return
        self.__append_to_queue(game)
        updates, json_data = self.__prepare_download(game, force)
        if not updates:
            self.__remove_from_queue(game)
            load_callback()
        else:
            # Copy the data because we are going to be a thread and we modify them later on
            image_worker = ImageManager.Worker(self.__download, updates.copy(), json_data.copy(), game)
            image_worker.signals.completed.connect(self.__remove_from_queue)
            image_worker.signals.completed.connect(load_callback)
            self.threadpool.start(image_worker, priority)

    def download_image_launch(
            self, game: Game, callback: Callable[[Game], None], priority: int, force: bool = False
    ) -> None:
        if self.__img_cache(game.app_name).is_file() and not force:
            return

        def _callback():
            callback(game)

        self.download_image(game, _callback, priority, force)

    def download_image_blocking(self, game: Game, force: bool = False) -> None:
        updates, json_data = self.__prepare_download(game, force)
        if not updates:
            return
        if updates:
            self.__download(updates, json_data, game, use_async=True)

    @staticmethod
    def __get_cover(
            container: Union[Type[QPixmap], Type[QImage]], app_name: str, preset: ImageSize.Preset, color: bool,
    ) -> Union[QPixmap, QImage]:
        ret = container()
        if preset.orientation == ImageType.Icon:
            if image_icon_path(app_name, color).is_file():
                ret.load(image_icon_path(app_name, color).as_posix())
        elif preset.orientation == ImageType.Tall:
            if image_tall_path(app_name, color).is_file():
                ret.load(image_tall_path(app_name, color).as_posix())
        elif preset.orientation == ImageType.Wide:
            if image_wide_path(app_name, color).is_file():
                ret.load(image_wide_path(app_name, color).as_posix())
        else:
            raise RuntimeError("Unknown image preset")
        if not ret.isNull():
            device = ImageSize.Preset(
                divisor=preset.base.divisor,
                pixel_ratio=QApplication.instance().devicePixelRatio(),
                orientation=preset.base.orientation,
                base=preset
            )
            ret.setDevicePixelRatio(preset.pixel_ratio)
            # lk: Scaling happens at painting. It might be inefficient so leave this here as an alternative
            # lk: If this is uncommented, the transformation in ImageWidget should be adjusted also
            ret = ret.scaled(
                device.size,
                Qt.AspectRatioMode.KeepAspectRatio,
                Qt.TransformationMode.SmoothTransformation
            )
            ret.setDevicePixelRatio(device.pixel_ratio)
        return ret

    def get_pixmap(self, app_name: str, preset: ImageSize.Preset, color: bool = True) -> QPixmap:
        """
        Use when the image is to be presented directly on the screen.

        @param app_name: The RareGame object for this game
        @param preset:
        @param color: True to load the colored pixmap, False to load the grayscale
        @return: QPixmap
        """
        pixmap: QPixmap = self.__get_cover(QPixmap, app_name, preset, color)
        return pixmap

    def get_image(self, app_name: str, preset: ImageSize.Preset, color: bool = True) -> QImage:
        """
        Use when the image has to be manipulated before being rendered.

        @param app_name: The RareGame object for this game
        @param preset:
        @param color: True to load the colored image, False to load the grayscale
        @return: QImage
        """
        image: QImage = self.__get_cover(QImage, app_name, preset, color)
        return image



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'import_login.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QFormLayout, QHBoxLayout,
    QLabel, QLayout, QPushButton, QSizePolicy,
    QSpacerItem, QVBoxLayout, QWidget)

class Ui_ImportLogin(object):
    def setupUi(self, ImportLogin):
        if not ImportLogin.objectName():
            ImportLogin.setObjectName(u"ImportLogin")
        ImportLogin.resize(256, 143)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(ImportLogin.sizePolicy().hasHeightForWidth())
        ImportLogin.setSizePolicy(sizePolicy)
        ImportLogin.setWindowTitle(u"ImportLogin")
        self.main_layout = QVBoxLayout(ImportLogin)
        self.main_layout.setObjectName(u"main_layout")
        self.title_label = QLabel(ImportLogin)
        self.title_label.setObjectName(u"title_label")
        sizePolicy.setHeightForWidth(self.title_label.sizePolicy().hasHeightForWidth())
        self.title_label.setSizePolicy(sizePolicy)
        font = QFont()
        font.setBold(True)
        self.title_label.setFont(font)

        self.main_layout.addWidget(self.title_label)

        self.form_layout = QFormLayout()
        self.form_layout.setObjectName(u"form_layout")
        self.form_layout.setSizeConstraint(QLayout.SetFixedSize)
        self.form_layout.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)
        self.form_layout.setLabelAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)
        self.form_layout.setFormAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignVCenter)
        self.prefix_layout = QHBoxLayout()
        self.prefix_layout.setObjectName(u"prefix_layout")
        self.prefix_combo = QComboBox(ImportLogin)
        self.prefix_combo.setObjectName(u"prefix_combo")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.prefix_combo.sizePolicy().hasHeightForWidth())
        self.prefix_combo.setSizePolicy(sizePolicy1)
        self.prefix_combo.setEditable(True)

        self.prefix_layout.addWidget(self.prefix_combo)

        self.prefix_button = QPushButton(ImportLogin)
        self.prefix_button.setObjectName(u"prefix_button")

        self.prefix_layout.addWidget(self.prefix_button)

        self.prefix_layout.setStretch(0, 1)

        self.form_layout.setLayout(0, QFormLayout.ItemRole.FieldRole, self.prefix_layout)

        self.prefix_label = QLabel(ImportLogin)
        self.prefix_label.setObjectName(u"prefix_label")

        self.form_layout.setWidget(0, QFormLayout.LabelRole, self.prefix_label)

        self.status_label = QLabel(ImportLogin)
        self.status_label.setObjectName(u"status_label")
        font1 = QFont()
        font1.setItalic(True)
        self.status_label.setFont(font1)
        self.status_label.setText(u"")

        self.form_layout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.status_label)


        self.main_layout.addLayout(self.form_layout)

        self.vspacer = QSpacerItem(20, 10, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.main_layout.addItem(self.vspacer)

        self.info_label = QLabel(ImportLogin)
        self.info_label.setObjectName(u"info_label")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.info_label.sizePolicy().hasHeightForWidth())
        self.info_label.setSizePolicy(sizePolicy2)
        self.info_label.setWordWrap(True)

        self.main_layout.addWidget(self.info_label)


        self.retranslateUi(ImportLogin)
    # setupUi

    def retranslateUi(self, ImportLogin):
        self.title_label.setText(QCoreApplication.translate("ImportLogin", u"Import existing session from EGL", None))
        self.prefix_button.setText(QCoreApplication.translate("ImportLogin", u"Browse...", None))
        self.prefix_label.setText(QCoreApplication.translate("ImportLogin", u"Select prefix", None))
        self.info_label.setText(QCoreApplication.translate("ImportLogin", u"<i>Please select the Wine prefix where Epic Games Launcher is installed. You will get logged out from EGL in the process.</i>", None))
        pass
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'download_widget.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QGridLayout, QLabel, QProgressBar,
    QPushButton, QSizePolicy, QWidget)

class Ui_DownloadWidget(object):
    def setupUi(self, DownloadWidget):
        if not DownloadWidget.objectName():
            DownloadWidget.setObjectName(u"DownloadWidget")
        DownloadWidget.resize(479, 86)
        DownloadWidget.setWindowTitle(u"DownloadWidget")
        self.active_download_layout = QGridLayout(DownloadWidget)
        self.active_download_layout.setObjectName(u"active_download_layout")
        self.active_download_layout.setContentsMargins(0, 0, 0, 0)
        self.dl_name = QLabel(DownloadWidget)
        self.dl_name.setObjectName(u"dl_name")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.dl_name.sizePolicy().hasHeightForWidth())
        self.dl_name.setSizePolicy(sizePolicy)
        font = QFont()
        font.setBold(False)
        self.dl_name.setFont(font)
        self.dl_name.setText(u"dl_name")

        self.active_download_layout.addWidget(self.dl_name, 0, 1, 1, 3)

        self.dl_name_label = QLabel(DownloadWidget)
        self.dl_name_label.setObjectName(u"dl_name_label")
        font1 = QFont()
        font1.setBold(True)
        self.dl_name_label.setFont(font1)
        self.dl_name_label.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)

        self.active_download_layout.addWidget(self.dl_name_label, 0, 0, 1, 1)

        self.cache_used = QLabel(DownloadWidget)
        self.cache_used.setObjectName(u"cache_used")
        sizePolicy.setHeightForWidth(self.cache_used.sizePolicy().hasHeightForWidth())
        self.cache_used.setSizePolicy(sizePolicy)
        self.cache_used.setText(u"cache_used")

        self.active_download_layout.addWidget(self.cache_used, 2, 1, 1, 1)

        self.dl_speed_label = QLabel(DownloadWidget)
        self.dl_speed_label.setObjectName(u"dl_speed_label")
        self.dl_speed_label.setFont(font1)
        self.dl_speed_label.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)

        self.active_download_layout.addWidget(self.dl_speed_label, 1, 2, 1, 1)

        self.downloaded_label = QLabel(DownloadWidget)
        self.downloaded_label.setObjectName(u"downloaded_label")
        self.downloaded_label.setFont(font1)
        self.downloaded_label.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)

        self.active_download_layout.addWidget(self.downloaded_label, 2, 2, 1, 1)

        self.dl_speed = QLabel(DownloadWidget)
        self.dl_speed.setObjectName(u"dl_speed")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.dl_speed.sizePolicy().hasHeightForWidth())
        self.dl_speed.setSizePolicy(sizePolicy1)
        self.dl_speed.setText(u"dl_speed")

        self.active_download_layout.addWidget(self.dl_speed, 1, 3, 1, 1)

        self.time_left_label = QLabel(DownloadWidget)
        self.time_left_label.setObjectName(u"time_left_label")
        self.time_left_label.setFont(font1)
        self.time_left_label.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)

        self.active_download_layout.addWidget(self.time_left_label, 1, 0, 1, 1)

        self.downloaded = QLabel(DownloadWidget)
        self.downloaded.setObjectName(u"downloaded")
        sizePolicy1.setHeightForWidth(self.downloaded.sizePolicy().hasHeightForWidth())
        self.downloaded.setSizePolicy(sizePolicy1)
        self.downloaded.setText(u"downloaded")

        self.active_download_layout.addWidget(self.downloaded, 2, 3, 1, 1)

        self.time_left = QLabel(DownloadWidget)
        self.time_left.setObjectName(u"time_left")
        sizePolicy.setHeightForWidth(self.time_left.sizePolicy().hasHeightForWidth())
        self.time_left.setSizePolicy(sizePolicy)
        self.time_left.setText(u"time_left")

        self.active_download_layout.addWidget(self.time_left, 1, 1, 1, 1)

        self.cache_used_label = QLabel(DownloadWidget)
        self.cache_used_label.setObjectName(u"cache_used_label")
        self.cache_used_label.setFont(font1)
        self.cache_used_label.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)

        self.active_download_layout.addWidget(self.cache_used_label, 2, 0, 1, 1)

        self.progress_bar = QProgressBar(DownloadWidget)
        self.progress_bar.setObjectName(u"progress_bar")
        self.progress_bar.setValue(0)

        self.active_download_layout.addWidget(self.progress_bar, 3, 0, 1, 4)

        self.kill_button = QPushButton(DownloadWidget)
        self.kill_button.setObjectName(u"kill_button")

        self.active_download_layout.addWidget(self.kill_button, 3, 4, 1, 1)

        self.active_download_layout.setColumnStretch(1, 1)
        self.active_download_layout.setColumnStretch(3, 1)

        self.retranslateUi(DownloadWidget)
    # setupUi

    def retranslateUi(self, DownloadWidget):
        self.dl_name_label.setText(QCoreApplication.translate("DownloadWidget", u"Current download:", None))
        self.dl_speed_label.setText(QCoreApplication.translate("DownloadWidget", u"Download speed:", None))
        self.downloaded_label.setText(QCoreApplication.translate("DownloadWidget", u"Downloaded:", None))
        self.time_left_label.setText(QCoreApplication.translate("DownloadWidget", u"Time left:", None))
        self.cache_used_label.setText(QCoreApplication.translate("DownloadWidget", u"Cache size:", None))
        self.kill_button.setText(QCoreApplication.translate("DownloadWidget", u"Stop Download", None))
        pass
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'egl_sync_list_group.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QGroupBox, QHBoxLayout, QLabel,
    QListWidget, QListWidgetItem, QPushButton, QSizePolicy,
    QSpacerItem, QVBoxLayout, QWidget)

class Ui_EGLSyncListGroup(object):
    def setupUi(self, EGLSyncListGroup):
        if not EGLSyncListGroup.objectName():
            EGLSyncListGroup.setObjectName(u"EGLSyncListGroup")
        EGLSyncListGroup.resize(461, 206)
        EGLSyncListGroup.setWindowTitle(u"EGLSyncListGroup")
        EGLSyncListGroup.setTitle(u"")
        self.egl_sync_list_layout = QVBoxLayout(EGLSyncListGroup)
        self.egl_sync_list_layout.setObjectName(u"egl_sync_list_layout")
        self.label = QLabel(EGLSyncListGroup)
        self.label.setObjectName(u"label")
        self.label.setText(u"")

        self.egl_sync_list_layout.addWidget(self.label)

        self.list = QListWidget(EGLSyncListGroup)
        self.list.setObjectName(u"list")
        self.list.setAlternatingRowColors(True)
        self.list.setSpacing(1)
        self.list.setSortingEnabled(True)

        self.egl_sync_list_layout.addWidget(self.list)

        self.buttons_widget = QWidget(EGLSyncListGroup)
        self.buttons_widget.setObjectName(u"buttons_widget")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttons_widget.sizePolicy().hasHeightForWidth())
        self.buttons_widget.setSizePolicy(sizePolicy)
        self.buttons_layout = QHBoxLayout(self.buttons_widget)
        self.buttons_layout.setObjectName(u"buttons_layout")
        self.buttons_layout.setContentsMargins(0, 0, 0, 0)
        self.select_all_button = QPushButton(self.buttons_widget)
        self.select_all_button.setObjectName(u"select_all_button")

        self.buttons_layout.addWidget(self.select_all_button)

        self.select_none_button = QPushButton(self.buttons_widget)
        self.select_none_button.setObjectName(u"select_none_button")

        self.buttons_layout.addWidget(self.select_none_button)

        self.buttons_hspacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.buttons_layout.addItem(self.buttons_hspacer)

        self.action_button = QPushButton(self.buttons_widget)
        self.action_button.setObjectName(u"action_button")
        self.action_button.setText(u"Action")

        self.buttons_layout.addWidget(self.action_button)


        self.egl_sync_list_layout.addWidget(self.buttons_widget)


        self.retranslateUi(EGLSyncListGroup)
    # setupUi

    def retranslateUi(self, EGLSyncListGroup):
        self.select_all_button.setText(QCoreApplication.translate("EGLSyncListGroup", u"Select all", None))
        self.select_none_button.setText(QCoreApplication.translate("EGLSyncListGroup", u"Select none", None))
        pass
    # retranslateUi




import os
import shlex
from dataclasses import dataclass
from enum import Enum
from hashlib import md5
from logging import getLogger
from typing import Optional, Union, List, Dict, Set

import vdf

logger = getLogger("SteamTools")

steam_client_install_paths = [os.path.expanduser("~/.local/share/Steam")]


def find_steam() -> Optional[str]:
    # return the first valid path
    for path in steam_client_install_paths:
        if os.path.isdir(path) and os.path.isfile(os.path.join(path, "steam.sh")):
            return path
    return None


def find_libraries(steam_path: str) -> Set[str]:
    vdf_path = os.path.join(steam_path, "config", "libraryfolders.vdf")
    with open(vdf_path, "r", encoding="utf-8") as f:
        libraryfolders = vdf.load(f)["libraryfolders"]
    # libraries = [os.path.join(folder["path"], "steamapps") for key, folder in libraryfolders.items()]
    libraries = {os.path.join(folder["path"], "steamapps") for key, folder in libraryfolders.items()}
    libraries = set(filter(lambda x: os.path.isdir(x), libraries))
    return libraries


# Notes:
# Anything older than 'Proton 5.13' doesn't have the 'require_tool_appid' attribute.
# Anything older than 'Proton 7.0' doesn't have the 'compatmanager_layer_name' attribute.
# In addition to that, the 'Steam Linux Runtime 1.0 (scout)' runtime lists the
# 'Steam Linux Runtime 2.0 (soldier)' runtime as a dependency and is probably what was
# being used for any version before 5.13.
#
# As a result the following implementation will list versions from 7.0 onwards which honestly
# is a good trade-off for the amount of complexity supporting everything would ensue.


class SteamVerb(Enum):
    RUN = "run"
    WAIT_FOR_EXIT_AND_RUN = "waitforexitandrun"
    RUN_IN_PREFIX = "runinprefix"
    DESTROY_PREFIX = "destroyprefix"
    GET_COMPAT_PATH = "getcompatpath"
    GET_NATIVE_PATH = "getnativepath"
    DEFAULT = WAIT_FOR_EXIT_AND_RUN


@dataclass
class SteamBase:
    steam_path: str
    tool_path: str
    toolmanifest: Dict

    def __eq__(self, other):
        return self.tool_path == other.tool_path

    def __hash__(self):
        return hash(self.tool_path)

    @property
    def required_tool(self) -> Optional[str]:
        return self.toolmanifest.get("require_tool_appid", None)

    @property
    def layer(self) -> Optional[str]:
        return self.toolmanifest.get("compatmanager_layer_name", None)

    def command(self, verb: SteamVerb = SteamVerb.DEFAULT) -> List[str]:
        tool_path = os.path.normpath(self.tool_path)
        cmd = "".join([shlex.quote(tool_path), self.toolmanifest["commandline"]])
        # NOTE: "waitforexitandrun" seems to be the verb used in by steam to execute stuff
        # `run` is used when setting up the environment, so use that if we are setting up the prefix.
        cmd = cmd.replace("%verb%", verb.value)
        return shlex.split(cmd)

    def as_str(self, verb: SteamVerb = SteamVerb.DEFAULT):
        return " ".join(map(shlex.quote, self.command(verb)))

    @property
    def checksum(self) -> str:
        return md5(self.as_str().encode("utf-8")).hexdigest()


@dataclass
class SteamRuntime(SteamBase):
    steam_library: str
    appmanifest: Dict

    @property
    def name(self) -> str:
        return self.appmanifest["AppState"]["name"]

    @property
    def appid(self) -> str:
        return self.appmanifest["AppState"]["appid"]


@dataclass
class SteamAntiCheat:
    steam_path: str
    tool_path: str
    steam_library: str
    appmanifest: Dict

    def __eq__(self, other):
        return self.tool_path == other.tool_path

    def __hash__(self):
        return hash(self.tool_path)

    @property
    def name(self) -> str:
        return self.appmanifest["AppState"]["name"]

    @property
    def appid(self) -> str:
        return self.appmanifest["AppState"]["appid"]


@dataclass
class ProtonTool(SteamRuntime):
    runtime: SteamRuntime = None
    anticheat: Dict[str, SteamAntiCheat] = None

    def __bool__(self) -> bool:
        if appid := self.required_tool:
            return self.runtime is not None and self.runtime.appid == appid
        return True

    def command(self, verb: SteamVerb = SteamVerb.DEFAULT) -> List[str]:
        cmd = self.runtime.command(verb)
        cmd.extend(super().command(verb))
        return cmd


@dataclass
class CompatibilityTool(SteamBase):
    compatibilitytool: Dict
    runtime: SteamRuntime = None
    anticheat: Dict[str, SteamAntiCheat] = None

    def __bool__(self) -> bool:
        if appid := self.required_tool:
            return self.runtime is not None and self.runtime.appid == appid
        return True

    @property
    def name(self) -> str:
        return self.compatibilitytool["display_name"]

    def command(self, verb: SteamVerb = SteamVerb.DEFAULT) -> List[str]:
        cmd = self.runtime.command(verb) if self.runtime is not None else []
        cmd.extend(super().command(verb))
        return cmd


def find_appmanifests(library: str) -> List[dict]:
    appmanifests = []
    for entry in os.scandir(library):
        if entry.is_file() and entry.name.endswith(".acf"):
            with open(os.path.join(library, entry.name), "r", encoding="utf-8") as f:
                appmanifest = vdf.load(f)
            appmanifests.append(appmanifest)
    return appmanifests


ANTICHEAT_RUNTIMES = {
    "eac_runtime": "1826330",
    "battleye_runtime": "1161040",
}


def find_anticheat(steam_path: str, library: str):
    runtimes = {}
    appmanifests = find_appmanifests(library)
    common = os.path.join(library, "common")
    for appmanifest in appmanifests:
        if appmanifest["AppState"]["appid"] not in ANTICHEAT_RUNTIMES.values():
            continue
        folder = appmanifest["AppState"]["installdir"]
        runtimes.update(
            {
                appmanifest["AppState"]["appid"]: SteamAntiCheat(
                    steam_path=steam_path,
                    steam_library=library,
                    appmanifest=appmanifest,
                    tool_path=os.path.join(common, folder),
                )
            }
        )
    return runtimes


def find_runtimes(steam_path: str, library: str) -> Dict[str, SteamRuntime]:
    runtimes = {}
    appmanifests = find_appmanifests(library)
    common = os.path.join(library, "common")
    for appmanifest in appmanifests:
        folder = appmanifest["AppState"]["installdir"]
        tool_path = os.path.join(common, folder)
        if os.path.isfile(vdf_file := os.path.join(tool_path, "toolmanifest.vdf")):
            with open(vdf_file, "r", encoding="utf-8") as f:
                toolmanifest = vdf.load(f)
                if toolmanifest["manifest"]["compatmanager_layer_name"] == "container-runtime":
                    runtimes.update(
                        {
                            appmanifest["AppState"]["appid"]: SteamRuntime(
                                steam_path=steam_path,
                                steam_library=library,
                                appmanifest=appmanifest,
                                tool_path=tool_path,
                                toolmanifest=toolmanifest["manifest"],
                            )
                        }
                    )
    return runtimes


def find_protons(steam_path: str, library: str) -> List[ProtonTool]:
    protons = []
    appmanifests = find_appmanifests(library)
    common = os.path.join(library, "common")
    for appmanifest in appmanifests:
        folder = appmanifest["AppState"]["installdir"]
        tool_path = os.path.join(common, folder)
        if os.path.isfile(vdf_file := os.path.join(tool_path, "toolmanifest.vdf")):
            with open(vdf_file, "r", encoding="utf-8") as f:
                toolmanifest = vdf.load(f)
                if toolmanifest["manifest"]["compatmanager_layer_name"] == "proton":
                    protons.append(
                        ProtonTool(
                            steam_path=steam_path,
                            steam_library=library,
                            appmanifest=appmanifest,
                            tool_path=tool_path,
                            toolmanifest=toolmanifest["manifest"],
                        )
                    )
    return protons


def find_compatibility_tools(steam_path: str) -> List[CompatibilityTool]:
    compatibilitytools_paths = {
        "/usr/share/steam/compatibilitytools.d",
        os.path.expanduser(os.path.join(steam_path, "compatibilitytools.d")),
        os.path.expanduser("~/.steam/compatibilitytools.d"),
        os.path.expanduser("~/.steam/root/compatibilitytools.d"),
    }
    compatibilitytools_paths = {
        os.path.realpath(path) for path in compatibilitytools_paths if os.path.isdir(path)
    }
    tools = []
    for path in compatibilitytools_paths:
        for entry in os.scandir(path):
            entry_path = os.path.join(path, entry.name)
            if entry.is_dir():
                tool_vdf = os.path.join(entry_path, "compatibilitytool.vdf")
            elif entry.is_file() and entry.name.endswith(".vdf"):
                tool_vdf = entry_path
            else:
                continue

            if not os.path.isfile(tool_vdf):
                continue

            with open(tool_vdf, "r", encoding="utf-8") as f:
                compatibilitytool = vdf.load(f)

            entry_tools = compatibilitytool["compatibilitytools"]["compat_tools"]
            for entry_tool in entry_tools.values():
                if entry_tool["from_oslist"] != "windows" and entry_tool["to_oslist"] != "linux":
                    continue

                install_path = entry_tool["install_path"]
                tool_path = os.path.abspath(os.path.join(os.path.dirname(tool_vdf), install_path))
                manifest_vdf = os.path.join(tool_path, "toolmanifest.vdf")

                if not os.path.isfile(manifest_vdf):
                    continue

                with open(manifest_vdf, "r", encoding="utf-8") as f:
                    manifest = vdf.load(f)

                tools.append(
                    CompatibilityTool(
                        steam_path=steam_path,
                        tool_path=tool_path,
                        toolmanifest=manifest["manifest"],
                        compatibilitytool=entry_tool,
                    )
                )
    return tools


def get_runtime(
    tool: Union[ProtonTool, CompatibilityTool], runtimes: Dict[str, SteamRuntime]
) -> Optional[SteamRuntime]:
    required_tool = tool.required_tool
    if required_tool is None:
        return None
    return runtimes.get(required_tool, None)


def get_umu_environment(
    tool: Optional[ProtonTool] = None, compat_path: Optional[str] = None
) -> Dict:
    # If the tool is unset, return all affected env variable names
    # IMPORTANT: keep this in sync with the code below
    environ = {"WINEPREFIX": compat_path if compat_path else ""}
    if tool is None:
        environ["WINEPREFIX"] = ""
        environ["PROTONPATH"] = ""
        environ["GAMEID"] = ""
        environ["STORE"] = ""
        return environ

    environ["PROTONPATH"] = tool.tool_path
    environ["STORE"] = "egs"
    return environ


def get_steam_environment(
    tool: Optional[Union[ProtonTool, CompatibilityTool]] = None, compat_path: Optional[str] = None
) -> Dict:
    # If the tool is unset, return all affected env variable names
    # IMPORTANT: keep this in sync with the code below
    environ = {"STEAM_COMPAT_DATA_PATH": compat_path if compat_path else ""}
    if tool is None:
        environ["STEAM_COMPAT_DATA_PATH"] = ""
        environ["STEAM_COMPAT_CLIENT_INSTALL_PATH"] = ""
        environ["STEAM_COMPAT_LIBRARY_PATHS"] = ""
        environ["STEAM_COMPAT_MOUNTS"] = ""
        environ["STEAM_COMPAT_TOOL_PATHS"] = ""
        environ["PROTON_EAC_RUNTIME"] = ""
        environ["PROTON_BATTLEYE_RUNTIME"] = ""
        return environ

    environ["STEAM_COMPAT_CLIENT_INSTALL_PATH"] = tool.steam_path
    if isinstance(tool, ProtonTool):
        environ["STEAM_COMPAT_LIBRARY_PATHS"] = tool.steam_library
    if tool.runtime is not None:
        compat_mounts = [tool.tool_path, tool.runtime.tool_path]
        environ["STEAM_COMPAT_MOUNTS"] = ":".join(compat_mounts)
    tool_paths = [tool.tool_path]
    if tool.runtime is not None:
        tool_paths.append(tool.runtime.tool_path)
    environ["STEAM_COMPAT_TOOL_PATHS"] = ":".join(tool_paths)
    if tool.anticheat is not None:
        if (appid := ANTICHEAT_RUNTIMES["eac_runtime"]) in tool.anticheat.keys():
            environ["PROTON_EAC_RUNTIME"] = tool.anticheat[appid].tool_path
        if (appid := ANTICHEAT_RUNTIMES["battleye_runtime"]) in tool.anticheat.keys():
            environ["PROTON_BATTLEYE_RUNTIME"] = tool.anticheat[appid].tool_path
    return environ


def _find_tools() -> List[Union[ProtonTool, CompatibilityTool]]:
    steam_path = find_steam()
    if steam_path is None:
        logger.info("Steam could not be found")
        return []
    logger.info("Found Steam in %s", steam_path)

    steam_libraries = find_libraries(steam_path)
    logger.debug("Searching for tools in libraries:")
    logger.debug("%s", steam_libraries)

    runtimes = {}
    for library in steam_libraries:
        runtimes.update(find_runtimes(steam_path, library))

    anticheat = {}
    for library in steam_libraries:
        anticheat.update(find_anticheat(steam_path, library))

    tools = []
    for library in steam_libraries:
        tools.extend(find_protons(steam_path, library))
    tools.extend(find_compatibility_tools(steam_path))

    for tool in tools:
        runtime = get_runtime(tool, runtimes)
        tool.runtime = runtime
        if tool.layer == "proton":
            tool.anticheat = anticheat

    tools = list(filter(lambda t: bool(t), tools))

    return tools


_tools: Optional[List[Union[ProtonTool, CompatibilityTool]]] = None


def find_tools() -> List[Union[ProtonTool, CompatibilityTool]]:
    global _tools
    if _tools is None:
        _tools = _find_tools()
    return list(filter(lambda t: t.layer != "umu-launcher", _tools))


def find_umu_launcher() -> Optional[CompatibilityTool]:
    global _tools
    if _tools is None:
        _tools = _find_tools()
    _umu = list(filter(lambda t: t.layer == "umu-launcher",  _tools))
    return _umu[0] if _umu else None


if __name__ == "__main__":
    from pprint import pprint

    tools = find_tools()
    for tool in tools:
        pprint(tool)
    umu = find_umu_launcher()
    pprint(umu)


    for tool in tools:
        print(get_steam_environment(tool))
        print(tool.name)
        print(tool.command(SteamVerb.RUN))
        print(" ".join(tool.command(SteamVerb.RUN_IN_PREFIX)))



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'landing_page.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFormLayout, QLabel, QRadioButton,
    QSizePolicy, QWidget)

class Ui_LandingPage(object):
    def setupUi(self, LandingPage):
        if not LandingPage.objectName():
            LandingPage.setObjectName(u"LandingPage")
        LandingPage.resize(293, 78)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(LandingPage.sizePolicy().hasHeightForWidth())
        LandingPage.setSizePolicy(sizePolicy)
        LandingPage.setWindowTitle(u"LandingPage")
        self.main_layout = QFormLayout(LandingPage)
        self.main_layout.setObjectName(u"main_layout")
        self.main_layout.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)
        self.main_layout.setLabelAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignVCenter)
        self.main_layout.setFormAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignVCenter)
        self.login_label = QLabel(LandingPage)
        self.login_label.setObjectName(u"login_label")
        sizePolicy.setHeightForWidth(self.login_label.sizePolicy().hasHeightForWidth())
        self.login_label.setSizePolicy(sizePolicy)
        font = QFont()
        font.setBold(True)
        self.login_label.setFont(font)

        self.main_layout.setWidget(0, QFormLayout.SpanningRole, self.login_label)

        self.login_browser_radio = QRadioButton(LandingPage)
        self.login_browser_radio.setObjectName(u"login_browser_radio")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.login_browser_radio.sizePolicy().hasHeightForWidth())
        self.login_browser_radio.setSizePolicy(sizePolicy1)

        self.main_layout.setWidget(1, QFormLayout.LabelRole, self.login_browser_radio)

        self.login_browser_label = QLabel(LandingPage)
        self.login_browser_label.setObjectName(u"login_browser_label")
        font1 = QFont()
        font1.setItalic(True)
        self.login_browser_label.setFont(font1)

        self.main_layout.setWidget(1, QFormLayout.ItemRole.FieldRole, self.login_browser_label)

        self.login_import_radio = QRadioButton(LandingPage)
        self.login_import_radio.setObjectName(u"login_import_radio")
        sizePolicy1.setHeightForWidth(self.login_import_radio.sizePolicy().hasHeightForWidth())
        self.login_import_radio.setSizePolicy(sizePolicy1)

        self.main_layout.setWidget(2, QFormLayout.LabelRole, self.login_import_radio)

        self.login_import_label = QLabel(LandingPage)
        self.login_import_label.setObjectName(u"login_import_label")
        self.login_import_label.setFont(font1)

        self.main_layout.setWidget(2, QFormLayout.ItemRole.FieldRole, self.login_import_label)


        self.retranslateUi(LandingPage)
    # setupUi

    def retranslateUi(self, LandingPage):
        self.login_label.setText(QCoreApplication.translate("LandingPage", u"Select login method", None))
        self.login_browser_radio.setText(QCoreApplication.translate("LandingPage", u"Browser", None))
        self.login_browser_label.setText(QCoreApplication.translate("LandingPage", u"Login using a browser.", None))
        self.login_import_radio.setText(QCoreApplication.translate("LandingPage", u"Import", None))
        self.login_import_label.setText(QCoreApplication.translate("LandingPage", u"Import from Epic Games Launcher", None))
        pass
    # retranslateUi

